#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU OVPSimCOREV Target definitions"""

import os
import re
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from mlonmcu.timeout import exec_timeout
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.bench import add_bench_metrics
from .riscv import RISCVTarget, sort_extensions_canonical

logger = get_logger()


MAX_P_SPEC = 0.96


def replace_unsupported(exts):
    ret = []
    for ext in exts:
        if "ZVE" in ext or "ZVL" in ext:
            ret.append("V")
        if "zpn" in ext.lower() or "zpsfoperand" in ext.lower():
            pass
        elif ext == "P":
            ret.append(ext)
            ret.append("B")
        else:
            ret.append(ext)
    return ret


class COREVOVPSimTarget(RISCVTarget):
    """TODO"""

    FEATURES = RISCVTarget.FEATURES | {"xcorev", "gdbserver", "log_instrs", "trace"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "variant": None,
        "processor": None,
        "fpu": "none",
        "atomic": False,
        "end_to_end_cycles": False,
        "gdbserver_enable": False,
        "gdbserver_attach": False,
        "gdbserver_port": 2222,
        "enable_xcorevmac": False,
        "enable_xcorevmem": False,
        "enable_xcorevbi": False,
        "enable_xcorevalu": False,
        "enable_xcorevbitmanip": False,
        "enable_xcorevsimd": False,
        "enable_xcorevhwlp": False,
    }
    REQUIRED = RISCVTarget.REQUIRED | {"corev_ovpsim.exe"}

    def __init__(self, name="corev_ovpsim", features=None, config=None):
        super().__init__(name, features=features, config=config)

    @property
    def ovpsim_exe(self):
        return Path(self.config["corev_ovpsim.exe"])

    @property
    def variant(self):
        temp = self.config["variant"]
        if temp:
            return temp
        else:
            return "CV32E40P"

    @property
    def processor(self):
        temp = self.config["processor"]
        if temp:
            return temp
        else:
            return "CVE4P"

    @property
    def enable_xcorevmac(self):
        value = self.config["enable_xcorevmac"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevmem(self):
        value = self.config["enable_xcorevmem"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbi(self):
        value = self.config["enable_xcorevbi"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevalu(self):
        value = self.config["enable_xcorevalu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbitmanip(self):
        value = self.config["enable_xcorevbitmanip"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevsimd(self):
        value = self.config["enable_xcorevsimd"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevhwlp(self):
        value = self.config["enable_xcorevhwlp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def extensions(self):
        exts = super().extensions
        required = set()
        if "xcorev" not in exts:
            if self.enable_xcorevmac:
                required.add("xcvmac")
            if self.enable_xcorevmem:
                required.add("xcvmem")
            if self.enable_xcorevbi:
                required.add("xcvbi")
            if self.enable_xcorevalu:
                required.add("xcvalu")
            if self.enable_xcorevbitmanip:
                required.add("xcvbitmanip")
            if self.enable_xcorevsimd:
                required.add("xcvsimd")
            if self.enable_xcorevhwlp:
                required.add("xcvhwlp")
        for ext in required:
            if ext not in exts:
                exts.add(ext)
        return exts

    @property
    def attr(self):
        attrs = super().attr.split(",")
        if self.enable_xcorevmac:
            if "xcorevmac" not in attrs:
                attrs.append("+xcvmac")
        if self.enable_xcorevmem:
            if "xcorevmem" not in attrs:
                attrs.append("+xcvmem")
        if self.enable_xcorevbi:
            if "xcorevbi" not in attrs:
                attrs.append("+xcvbi")
        if self.enable_xcorevalu:
            if "xcorevalu" not in attrs:
                attrs.append("+xcvalu")
        if self.enable_xcorevbitmanip:
            if "xcorevbitmanip" not in attrs:
                attrs.append("+xcvbitmanip")
        if self.enable_xcorevsimd:
            if "xcorevsimd" not in attrs:
                attrs.append("+xcvsimd")
        if self.enable_xcorevhwlp:
            if "xcorevhwlp" not in attrs:
                attrs.append("+xcvhwlp")
        return ",".join(attrs)

    @property
    def end_to_end_cycles(self):
        value = self.config["end_to_end_cycles"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_enable(self):
        value = self.config["gdbserver_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_attach(self):
        value = self.config["gdbserver_attach"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_port(self):
        return int(self.config["gdbserver_port"])

    def get_default_ovpsim_args(self):
        extensions_before = sort_extensions_canonical(self.extensions, lower=False, unpack=True)
        extensions_after = replace_unsupported(extensions_before)
        extensions_str = "".join(sort_extensions_canonical(extensions_after))
        args = [
            "--variant",
            self.variant,
            "--processorname",
            self.processor,
            "--override",
            f"riscvOVPsim/cpu/add_Extensions={extensions_str}",
            "--override",
            "riscvOVPsim/cpu/unaligned=T",
            "--override",
            "riscvOVPsim/cpu/pk/reportExitErrors=T",
            "--finishonopcode",
            "0",
            "--override",
            "riscvOVPsim/cpu/extension_CVE4P/mcountinhibit_reset=0",
        ]
        if self.gdbserver_enable:
            # args.append("--trace")
            args.extend(["--port", str(self.gdbserver_port)])
            if self.gdbserver_attach:
                args.append("--gdbconsole")
        return args

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        ovpsim_args = []

        ovpsim_args.extend(["--program", str(program)])
        ovpsim_args.extend(self.get_default_ovpsim_args())

        if len(self.extra_args) > 0:
            if isinstance(self.extra_args, str):
                extra_args = self.extra_args.split(" ")
            else:
                extra_args = self.extra_args
            ovpsim_args.extend(extra_args)  # I rename args to extra_args because otherwise it overwrites *args

        if self.timeout_sec > 0:
            ret = exec_timeout(
                self.timeout_sec,
                self.ovpsim_exe.resolve(),
                *ovpsim_args,
                *args,
                cwd=cwd,
                **kwargs,
            )
        else:
            ret = execute(
                self.ovpsim_exe.resolve(),
                *ovpsim_args,
                *args,  # Does this work?
                cwd=cwd,
                **kwargs,
            )
        return ret

    def parse_exit(self, out):
        exit_code = super().parse_exit(out)
        if exit_code is None:
            # legacy
            exit_code = None
            exit_match = re.search(r"Error \(RISCV/PK_EXIT\) Non-zero exit code: (.*)", out)
            if exit_match:
                exit_code = int(exit_match.group(1))
        return exit_code

    def parse_stdout(self, out, metrics, exit_code=0):
        add_bench_metrics(out, metrics, exit_code != 0, target_name=self.name)
        if self.end_to_end_cycles:
            sim_insns = re.search(r".*  Simulated instructions:(.*)", out)
            sim_insns = int(float(sim_insns.group(1)))
            metrics.add("Simulated Instructions", sim_insns, True)
        mips = None  # TODO: parse mips?
        mips_match = re.search(r".*  Simulated MIPS:(.*)", out)
        if mips_match:
            mips_str = float(mips_match.group(1))
            if "run too short for meaningful result" not in mips:
                mips = float(mips_str)
        if mips:
            metrics.add("MIPS", mips, optional=True)

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            # TODO: before or after?
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
        # TODO: get exit code
        exit_code = 0
        metrics = Metrics()
        self.parse_stdout(out, metrics, exit_code=exit_code)

        return metrics, out, []

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        return ret


if __name__ == "__main__":
    cli(target=COREVOVPSimTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU target submodule"""

from .target import Target
from ._target import register_target, get_targets
from .riscv import EtissPulpinoTarget, SpikeTarget, OVPSimTarget, RiscvQemuTarget
from .arm import Corstone300Target
from .host_x86 import HostX86Target

__all__ = [
    "register_target",
    "get_targets",
    "Target",
    "EtissPulpinoTarget",
    "SpikeTarget",
    "OVPSimTarget",
    "RiscvQemuTarget",
    "Corstone300Target",
    "HostX86Target",
]



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
from pathlib import Path


def make_hex_array(filename, mode="bin"):
    out = ""
    if mode == "auto":
        _, ext = os.path.splitext(filename)
        assert len(ext) > 1, "Could not detect format because of missing file extension"
        mode = ext[1:]
    if mode == "bin":
        with open(filename, "rb") as f:
            data = f.read(1)
            length = 0
            while data:
                length += 1
                out += "0x" + data.hex() + ", "
                data = f.read(1)
            assert length > 0, "Data can not be empty"
    elif mode in ["npy", "npz"]:
        data = np.load(filename)
        # TODO: figure out endianess
        if hasattr(data, "files"):
            files = data.files
            assert len(files) == 1
            data = data[files[0]]
        byte_data = data.tobytes()
        assert len(byte_data) > 0, "Data can not be empty"
        out = ", ".join(["0x{:02x}".format(x) for x in byte_data] + [""])
    else:
        raise RuntimeError(f"Unsupported mode: {mode}")
    return out


def fill_data_source(in_bufs, out_bufs):
    out = '#include "ml_interface.h"\n'
    out += "#include <stddef.h>\n"
    out += "const int num_data_buffers_in = " + str(sum([len(buf) for buf in in_bufs])) + ";\n"
    out += "const int num_data_buffers_out = " + str(sum([len(buf) for buf in out_bufs])) + ";\n"
    for i, buf in enumerate(in_bufs):
        for j in range(len(buf)):
            out += "const unsigned char data_buffer_in_" + str(i) + "_" + str(j) + "[] = {" + buf[j] + "};\n"
    for i, buf in enumerate(out_bufs):
        for j in range(len(buf)):
            out += "const unsigned char data_buffer_out_" + str(i) + "_" + str(j) + "[] = {" + buf[j] + "};\n"

    var_in = "const unsigned char *const data_buffers_in[] = {"
    var_insz = "const size_t data_size_in[] = {"
    for i, buf in enumerate(in_bufs):
        for j in range(len(buf)):
            var_in += "data_buffer_in_" + str(i) + "_" + str(j) + ", "
            var_insz += "sizeof(data_buffer_in_" + str(i) + "_" + str(j) + "), "
    var_out = "const unsigned char *const data_buffers_out[] = {"
    var_outsz = "const size_t data_size_out[] = {"
    for i, buf in enumerate(out_bufs):
        for j in range(len(buf)):
            var_out += "data_buffer_out_" + str(i) + "_" + str(j) + ", "
            var_outsz += "sizeof(data_buffer_out_" + str(i) + "_" + str(j) + "), "
    out += var_in + "};\n" + var_out + "};\n" + var_insz + "};\n" + var_outsz + "};\n"
    return out


def lookup_data_buffers(input_paths, output_paths):
    assert len(input_paths) > 0
    legacy = False
    used_fmt = None
    allowed_fmts = ["bin", "npy", "npz"]

    def helper(paths):
        nonlocal used_fmt, legacy
        data = []
        for i, path in enumerate(paths):
            if path.is_dir():
                filenames = os.listdir(path)
            else:
                filenames = [path]
            for filename in filenames:
                fmt = Path(filename).suffix[1:]
                if fmt not in allowed_fmts:
                    continue
                if used_fmt is None:
                    used_fmt = fmt
                else:
                    assert used_fmt == fmt, "Please only use a single format for inout model data (.bin OR .npy)"
                base = Path(filename).stem
                if "_" in base:
                    legacy = True
                    assert len(paths) == 1, "Legacy mode only allows a single path"
                    data_index, tensor_index = list(map(int, base.split("_")))[:2]
                else:
                    data_index, tensor_index = int(base), 0
                hex_data = make_hex_array(Path(path) / filename, mode=used_fmt)
                data.append((data_index, tensor_index, hex_data))
        sorted_data = sorted(data, key=lambda x: (x[0], x[1]))
        # TODO: get rid of this dirty workaround
        ret = []
        for a, b, c in sorted_data:
            if a >= len(ret):
                assert b == 0
                ret.append([c])
            else:
                ret[a].append(c)
        return ret
        # return [d[-1] for d in sorted_data]  # Extract last column

    ins = helper(input_paths)
    outs = helper(output_paths)
    assert len(ins) == len(outs)
    return ins, outs


def get_data_source(input_paths, output_paths):
    assert len(input_paths) == len(output_paths)
    if len(input_paths) == 0:
        return fill_data_source([], [])
    in_bufs, out_bufs = lookup_data_buffers(input_paths, output_paths)
    return fill_data_source(in_bufs, out_bufs)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TODO"""

import string
from datetime import datetime
from math import ceil, log2

# TODO: use this
# from tvm.relay.backend.utils import mangle_module_name


def calc_pages(workspace_size, page_size=2**10):
    # Determine the number of required pages
    assert workspace_size >= 0, "Workspace size can not be negative"
    crtPageSizeLog2 = log2(page_size)
    assert crtPageSizeLog2 == int(crtPageSizeLog2), "Page size has to be a power of two"
    crtNumPages = ceil(workspace_size / page_size)
    return int(crtNumPages), int(crtPageSizeLog2)


def generate_wrapper_header():
    out = """#ifndef TVM_WRAPPER_H
#define TVM_WRAPPER_H

#include <stddef.h>

int TVMWrap_Init();
void *TVMWrap_GetInputPtr(int index);
size_t TVMWrap_GetInputSize(int index);
size_t TVMWrap_GetNumInputs();
int TVMWrap_Run();
void *TVMWrap_GetOutputPtr(int index);
size_t TVMWrap_GetOutputSize(int index);
size_t TVMWrap_GetNumOutputs();

#endif  // TVM_WRAPPER_H
"""
    return out


def generate_header():
    time = datetime.now()
    header = f"""// This file is generated. Do not edit.
// Generated on: {time}
"""
    return header


def generate_common_includes():
    return """
#include <stdlib.h>
#include <stdarg.h>
#include <dlpack/dlpack.h>
#include "tvm/runtime/crt/error_codes.h"
#include "tvm/runtime/c_runtime_api.h"
#include "printing.h"
"""


def generate_graph_includes():
    out = generate_common_includes()
    out += """
#include "tvm/runtime/crt/packed_func.h"
#include "tvm/runtime/crt/crt.h"
#include "tvm/runtime/crt/graph_executor.h"
#include "tvm/runtime/crt/page_allocator.h"

"""
    return out


def generate_aot_includes(allocator):
    out = generate_common_includes()
    if allocator:
        out += '#include "tvm/runtime/crt/stack_allocator.h"\n'
    return out


def fill(template, **kwargs):
    return string.Template(template).substitute(**kwargs)


def getSizes(tensors):
    out = "size_t sizes[] = { "
    for t in tensors:
        out += str(t.size) + ", "
    out += "};"
    return out


def write_tvmrt_wrapper(path, graph, params, model_info, workspace_size):
    with open(path, "w") as f:
        text = generate_tvmrt_wrapper(graph, params, model_info, workspace_size)
        f.write(text)


def generate_tvmrt_wrapper(graph, params, model_info, workspace_size, debug_arena=False):
    crtNumPages, crtPageSizeLog2 = calc_pages(workspace_size)

    def escapeJson(j):
        return j.replace('"', '\\"').replace("\n", "\\\n")

    def toCArray(bin):
        result = ""
        for c in bin:
            result += hex(c) + ", "
        return result

    def getMeta(tensors, withNames=False):
        out = ""
        if withNames:
            out = "const char *names[] = { "
            for t in tensors:
                out += '"' + t.name + '", '
            out += "};\n    "

        out += "DLDataType dtypes[] = {"
        for t in tensors:
            if t.dtype == "float32":
                out += "{kDLFloat, 32, 1}"
            elif t.dtype == "uint8":
                out += "{kDLUInt, 8, 1}"
            elif t.dtype == "int8":
                out += "{kDLInt, 8, 1}"
            elif t.dtype == "uint64":
                out += "{kDLUInt, 64, 1}"
            elif t.dtype == "int64":
                out += "{kDLInt, 64, 1}"
            else:
                raise RuntimeError(f"Invalid type: {t.dtype}")
            out += ", "
        out += "};\n    "

        for i, t in enumerate(tensors):
            out += "int64_t shape_" + str(i) + "[] = { "
            for s in t.shape:
                out += str(s) + ", "
            out += "};\n    "
        out += "int64_t *shapes[] = { "
        for i, t in enumerate(tensors):
            out += "shape_" + str(i) + ", "
        out += "};\n"

        out += "    size_t ndims[] = { "
        for i, t in enumerate(tensors):
            out += str(len(t.shape)) + ", "
        out += "};\n    "

        for i, t in enumerate(tensors):
            out += "static uint8_t data_" + str(i) + "[" + str(t.size) + "];\n"
        out += "    uint8_t *data[] = { "
        for i, t in enumerate(tensors):
            out += "data_" + str(i) + ", "
        out += "};"

        return out

    out = ""
    out += generate_header()
    out += generate_graph_includes()
    out += 'const char * const g_graph = "' + escapeJson(graph) + '";\n'
    out += "const char g_params[] = { " + toCArray(params) + "\n};\n"
    out += "const uint64_t g_params_size = " + str(len(params)) + ";\n"

    mainCode = """


#define CRT_MEMORY_NUM_PAGES ${numPages}
#define CRT_MEMORY_PAGE_SIZE_LOG2 ${pageSizeLog2}
"""

    if debug_arena:  # This will enable the feature only if it is not overwritten by the user/compiler
        mainCode += """
#ifndef DEBUG_ARENA_USAGE
#define DEBUG_ARENA_USAGE 1
#endif
"""

    mainCode += """
#ifdef DEBUG_ARENA_USAGE
size_t max_arena_usage = 0;
#endif

static uint8_t g_crt_memory[CRT_MEMORY_NUM_PAGES * (1 << CRT_MEMORY_PAGE_SIZE_LOG2)];
static MemoryManagerInterface* g_memory_manager;

/*! \\brief macro to do C API call */
#define TVM_CCALL(func)                                                                   \\
    do {                                                                                  \\
        tvm_crt_error_t ret = (func);                                                     \\
        if (ret != kTvmErrorNoError) {                                                    \\
            TVMLogf("%s: %d: error: %s\\n", __FILE__, __LINE__, TVMGetLastError());       \\
            TVMPlatformAbort(ret);                                                        \\
        }                                                                                 \\
    } while (0)

TVMModuleHandle TVMArgs_AsModuleHandle(const TVMArgs* args, size_t index);

void __attribute__((noreturn)) TVMPlatformAbort(tvm_crt_error_t code)
{
    exit(1);
}

void TVMLogf(const char* msg, ...)
{
    va_list args;
    va_start(args, msg);
    DBGPRINTF(msg, args);
    va_end(args);
}

tvm_crt_error_t TVMPlatformMemoryAllocate(size_t num_bytes, DLDevice dev, void** out_ptr)
{
    tvm_crt_error_t ret = g_memory_manager->Allocate(g_memory_manager, num_bytes, dev, out_ptr);
#ifdef DEBUG_ARENA_USAGE
    // Use this to estimate the required number of pages
    // Run in DEBUG mode in insert value of MAX printed last into the following equation:
    // (This will round to the next power of 2 which might not be wanted!)
    // num_pages = 2**ceil(log2(MAX/page_size))
    size_t end = (size_t)(*out_ptr-(void*)g_crt_memory)+num_bytes;
    if (end > max_arena_usage)
    {
        max_arena_usage = end;
    }
#endif
    return ret;
}

tvm_crt_error_t TVMPlatformMemoryFree(void* ptr, DLDevice dev)
{
    return g_memory_manager->Free(g_memory_manager, ptr, dev);
}

tvm_crt_error_t TVMPlatformTimerStart()
{
    return kTvmErrorFunctionCallNotImplemented;
}

tvm_crt_error_t TVMPlatformTimerStop(double* elapsed_time_seconds)
{
  return kTvmErrorFunctionCallNotImplemented;
}

void *g_handle = NULL;

int TVMWrap_Init()
{
    int64_t device_type = kDLCPU;
    int64_t device_id = 0;

    TVMByteArray params;
    params.data = g_params;
    params.size = g_params_size;

    DLDevice dev;
    dev.device_type = (DLDeviceType)device_type;
    dev.device_id = device_id;

    // get pointers
    TVM_CCALL(PageMemoryManagerCreate(&g_memory_manager, g_crt_memory, sizeof(g_crt_memory),
                                      CRT_MEMORY_PAGE_SIZE_LOG2));
    TVM_CCALL(TVMInitializeRuntime());
    TVMPackedFunc pf;
    TVMArgs args = TVMArgs_Create(NULL, NULL, 0);
    TVM_CCALL(TVMPackedFunc_InitGlobalFunc(&pf, "runtime.SystemLib", &args));
    TVM_CCALL(TVMPackedFunc_Call(&pf));

    TVMModuleHandle mod_syslib = TVMArgs_AsModuleHandle(&pf.ret_value, 0);

    // run modules
    TVMGraphExecutor* graph_executor = NULL;
    TVM_CCALL(TVMGraphExecutor_Create(g_graph, mod_syslib, &dev, &graph_executor));
    TVM_CCALL(TVMGraphExecutor_LoadParams(graph_executor, params.data, params.size));

    //return graph_executor;
    g_handle = graph_executor;
    return 0;  // TODO
}

void *TVMWrap_GetInputPtr(int index)
{
    ${inMeta}

    DLTensor input;
    input.data = (void*)data[index];
    DLDevice device = {kDLCPU, 0};
    input.device = device;
    input.ndim = ndims[index];
    input.dtype = dtypes[index];
    input.shape = shapes[index];
    input.strides = NULL;
    input.byte_offset = 0;

    TVMGraphExecutor* graph_executor = (TVMGraphExecutor*)g_handle;
    TVMGraphExecutor_SetInput(graph_executor, names[index], &input);

    return data[index];
}

size_t TVMWrap_GetInputSize(int index)
{
    ${inSizes}

    return sizes[index];
}

size_t TVMWrap_GetNumInputs()
{
    return ${numInputs};
}

int TVMWrap_Run()
{
    TVMGraphExecutor* graph_executor = (TVMGraphExecutor*)g_handle;
    TVMGraphExecutor_Run(graph_executor);
#if DEBUG_ARENA_USAGE
    DBGPRINTF("\\nGraph executor arena max usage after model invocation: %lu bytes\\n", max_arena_usage);
#endif  // DEBUG_ARENA_USAGE
    return 0;  // TODO
}

void *TVMWrap_GetOutputPtr(int index)
{
    ${outMeta}

    DLTensor output;
    output.data = (void*)data[index];
    DLDevice device = {kDLCPU, 0};
    output.device = device;
    output.ndim = ndims[index];
    output.dtype = dtypes[index];
    output.shape = shapes[index];
    output.strides = NULL;
    output.byte_offset = 0;

    TVMGraphExecutor* graph_executor = (TVMGraphExecutor*)g_handle;
    TVMGraphExecutor_GetOutput(graph_executor, index, &output);

    return data[index];
}

size_t TVMWrap_GetOutputSize(int index)
{
    ${outSizes}

    return sizes[index];
}

size_t TVMWrap_GetNumOutputs()
{
    return ${numOutputs};
}
"""
    out += fill(
        mainCode,
        inMeta=getMeta(model_info.in_tensors, True),
        outMeta=getMeta(model_info.out_tensors),
        inSizes=getSizes(model_info.in_tensors),
        outSizes=getSizes(model_info.out_tensors),
        numInputs=len(model_info.in_tensors),
        numOutputs=len(model_info.out_tensors),
        numPages=crtNumPages,
        pageSizeLog2=crtPageSizeLog2,
    )
    return out


def generate_tvmaot_wrapper(model_info, workspace_size, mod_name, api="c", debug_arena=False):
    modPrefix = f"tvmgen_{mod_name}"

    def writeTensors(in_tensors, out_tensors, modPrefix, api):
        if api == "c":
            retStr = """
// Define data for input and output tensors
"""

            def writeTensorsHelper(tensors, prefix, out=False):
                lenTensors = len(tensors)
                direction = "out" if out else "in"
                ret = ""
                names = [f"{direction}put{i}_data" for i in range(lenTensors)]
                for i, t in enumerate(tensors):
                    ret += "char " + names[i] + "[" + str(t.size) + "];\n"
                ret += f"void* {direction}puts[] = {{" + ", ".join(names) + "};\n"
                ret += f"struct {prefix}_{direction}puts {prefix}_{direction}puts = {{" + "\n"
                for i, t in enumerate(tensors):
                    tensor_name = t.name.replace(":", "_").replace("/", "_").replace(".", "_").replace(";", "_")
                    ret += f"    .{tensor_name} = {names[i]}," + "\n"
                ret += "};\n"
                return ret

            retStr += writeTensorsHelper(in_tensors, modPrefix, False)
            retStr += writeTensorsHelper(out_tensors, modPrefix, True)
            return retStr
        elif api == "packed":
            retStr = """
// Define data for input and output tensors
"""

            def writeTensorsHelper(tensors, prefix, out=False):
                lenTensors = len(tensors)
                direction = "out" if out else "in"
                ret = ""
                names = [f"{direction}put{i}_data" for i in range(lenTensors)]
                for i, t in enumerate(tensors):
                    ret += "char " + names[i] + "[" + str(t.size) + "];\n"
                ret += f"void* {direction}puts[] = {{" + ", ".join(names) + "};\n"
                return ret

            retStr += writeTensorsHelper(in_tensors, modPrefix, False)
            retStr += writeTensorsHelper(out_tensors, modPrefix, True)
            return retStr
        else:
            raise RuntimeError("api has to be either 'c' or 'packed'")

    out = ""
    out += generate_header()
    includes = generate_aot_includes(workspace_size > 0)

    if api == "c":
        includes += '#include "${modPrefix}.h"\n'

    out += fill(includes, modPrefix=modPrefix)

    out += "\n"

    out += writeTensors(model_info.in_tensors, model_info.out_tensors, modPrefix, api)

    logging_code = """
void TVMLogf(const char* msg, ...)
{
    va_list args;
    va_start(args, msg);
    DBGPRINTF(msg, args);
    va_end(args);
}
"""

    out += logging_code

    if workspace_size > 0:
        workspace_code = """
#define WORKSPACE_SIZE (${workspaceBytes})
static uint8_t g_aot_memory[WORKSPACE_SIZE];
tvm_workspace_t app_workspace;
"""

        if debug_arena:  # This will enable the feature only if it is not overwritten by the user/compiler
            workspace_code += """
#ifndef DEBUG_ARENA_USAGE
#define DEBUG_ARENA_USAGE 1
#endif

#ifndef TVMAOT_DEBUG_ALLOCATIONS
#define TVMAOT_DEBUG_ALLOCATIONS 1
#endif
"""

        workspace_code += """
#ifdef DEBUG_ARENA_USAGE
size_t max_arena_usage = 0;
#endif

tvm_crt_error_t TVMPlatformMemoryAllocate(size_t num_bytes, DLDevice dev, void** out_ptr)
{
#ifdef TVMAOT_DEBUG_ALLOCATIONS
    if (num_bytes > (app_workspace.workspace + app_workspace.workspace_size - app_workspace.next_alloc))
    {
      TVMLogf("TVMPlatformMemoryAllocate(%lu): Allocation would overflow arena!\\n", num_bytes);
      return kTvmErrorPlatformNoMemory;
    }
#endif
    tvm_crt_error_t ret = StackMemoryManager_Allocate(&app_workspace, num_bytes, out_ptr);
#ifdef DEBUG_ARENA_USAGE
  // Use this to estimate the required number of bytes for the arena
  size_t end = app_workspace.next_alloc-app_workspace.workspace;
  if (end > max_arena_usage)
  {
    max_arena_usage = end;
  }
#endif
    return ret;
}
tvm_crt_error_t TVMPlatformMemoryFree(void* ptr, DLDevice dev)
{
#ifdef TVMAOT_DEBUG_ALLOCATIONS
    if ((uint8_t*)ptr < app_workspace.workspace || (uint8_t*)ptr >= app_workspace.next_alloc)
    {
      TVMLogf("TVMPlatformMemoryFree(%p): Invalid Memory region to be free'd!\\n", ptr);
      return kTvmErrorPlatformNoMemory;
    }
#endif
    return StackMemoryManager_Free(&app_workspace, ptr);
}
"""
        out += fill(workspace_code, workspaceBytes=int(workspace_size))
    else:
        workspace_code = """
tvm_crt_error_t TVMPlatformMemoryAllocate(size_t num_bytes, DLDevice dev, void** out_ptr)
{
    return kTvmErrorFunctionCallNotImplemented;
}
tvm_crt_error_t TVMPlatformMemoryFree(void* ptr, DLDevice dev)
{
    return kTvmErrorFunctionCallNotImplemented;
}
"""
        out += workspace_code

    mainCode = ""
    if api == "packed":
        mainCode += (
            "int32_t ${modPrefix}_run(void* args, void* type_code, int num_args,"
            + " void* out_value, void* out_type_code, void* resource_handle);\n"
        )

    mainCode += """
void __attribute__((noreturn)) TVMPlatformAbort(tvm_crt_error_t code)
{
    exit(1);
}

TVM_DLL int TVMFuncRegisterGlobal(const char* name, TVMFunctionHandle f, int override)
{
    return 0;
}

int TVMWrap_Init()
{
"""
    if workspace_size > 0:
        mainCode += "    StackMemoryManager_Init(&app_workspace, g_aot_memory, WORKSPACE_SIZE);"
    mainCode += """
    return 0;  // TODO
}

void *TVMWrap_GetInputPtr(int index)
{
    return inputs[index];
}

size_t TVMWrap_GetInputSize(int index)
{
    ${inSizes}

    return sizes[index];
}

size_t TVMWrap_GetNumInputs()
{
    return ${numInputs};
}

int TVMWrap_Run()
{"""
    if api == "c":
        mainCode += """
    int ret_val = ${modPrefix}_run(&${modPrefix}_inputs, &${modPrefix}_outputs);
    if (ret_val)
    {
        TVMPlatformAbort(kTvmErrorPlatformCheckFailure);
    }
"""
    elif api == "packed":
        mainCode += """
    static DLDevice fake_device = {kDLCPU, 0};
    static int64_t fake_dims = 0;
    static int64_t fake_shape = {0};

    DLTensor tensors[${numInputs} + ${numOutputs}];
    TVMValue values[${numInputs} + ${numOutputs}];
    int32_t typeids[${numInputs} + ${numOutputs}];

    for (size_t i = 0; i < ${numInputs}+${numOutputs}; i++)
    {
        tensors[i].device = fake_device;
        tensors[i].data = (i < ${numInputs}) ? inputs[i] : outputs[i - ${numInputs}];
        tensors[i].shape = &fake_shape;
        tensors[i].ndim = fake_dims;
        tensors[i].byte_offset = 0;
        tensors[i].strides = NULL;
        values[i].v_handle = &tensors[i];
    }

    int ret_val = ${modPrefix}_run(values, typeids, 0, NULL, 0, NULL);
    if (ret_val)
    {
        TVMPlatformAbort(kTvmErrorPlatformCheckFailure);
    }
    return 0;

"""
    else:
        raise RuntimeError("api can only be 'c' or 'packed'")

    if workspace_size > 0:
        mainCode += """
#if DEBUG_ARENA_USAGE
    DBGPRINTF("\\nAoT executor arena max usage after model invocation: %lu bytes\\n", max_arena_usage);
#endif  // DEBUG_ARENA_USAGE
"""
    mainCode += """
    return 0;  // TODO
}

void *TVMWrap_GetOutputPtr(int index)
{
    return outputs[index];
}

size_t TVMWrap_GetOutputSize(int index)
{
    ${outSizes}

    return sizes[index];
}

size_t TVMWrap_GetNumOutputs()
{
    return ${numOutputs};
}
"""
    out += fill(
        mainCode,
        inSizes=getSizes(model_info.in_tensors),
        outSizes=getSizes(model_info.out_tensors),
        numInputs=len(model_info.in_tensors),
        numOutputs=len(model_info.out_tensors),
        modPrefix=modPrefix,
    )
    return out


def write_tvmaot_wrapper(path, model_info, workspace_size, mod_name, api="c"):
    with open(path, "w") as f:
        text = generate_tvmaot_wrapper(model_info, workspace_size, mod_name, api=api)
        f.write(text)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import tempfile
import tarfile
from pathlib import Path
from typing import Tuple
import multiprocessing

from mlonmcu.flow.backend import Backend
from mlonmcu.setup import utils
from mlonmcu.timeout import exec_timeout
from mlonmcu.config import str2bool, str2list, str2dict
from mlonmcu.logging import get_logger
from .model_info import get_model_info, get_fallback_model_info, get_supported_formats, get_model_format
from mlonmcu.target.metrics import Metrics
from mlonmcu.artifact import Artifact, ArtifactFormat
from .python_utils import prepare_python_environment
from .tvmc_utils import (
    get_target_tvmc_args,
    get_pass_config_tvmc_args,
    get_disabled_pass_tvmc_args,
    get_runtime_executor_tvmc_args,
    get_input_shapes_tvmc_args,
    get_tuning_records_tvmc_args,
    get_desired_layout_args,
)

logger = get_logger()


class TVMBackend(Backend):
    registry = {}

    name = None

    FEATURES = {"autotuned", "cmsisnnbyoc", "muriscvnnbyoc", "disable_legalize", "moiopt", "uma_backends", "fuse_ops"}

    DEFAULTS = {
        "print_outputs": False,
        "opt_level": 3,
        "target_device": None,
        "target_mcpu": None,
        "target_march": None,
        "target_model": None,
        "target_mtriple": None,
        "target_mabi": None,
        "target_mattr": None,
        "target_keys": None,
        "target_num_cores": None,
        "extra_targets": None,  # list
        "extra_target_details": None,  # dict
        "desired_layout": None,  # optional: NCHW, NHWC, NHWC:HWOI, ...
        "desired_layout_ops": None,  # optional: conv2d, max_pool2d,...
        "desired_layout_map": None,  # optional, conv2d=NCHW, ...
        "disabled_passes": [],  # i.e. AlterOpLayout
        "extra_pass_config": {},  # TODO: some example (fuse_max_depth etc.)
        "use_tuning_results": False,
        "tvmc_extra_args": [],  # Currently compile subcommand only!
        "tvmc_custom_script": None,
        # See https://github.com/apache/tvm/blob/1115fd9bc261619ffa0539746ae0aebc46232dc6/python/tvm/autotvm/tophub.py
        "tophub_url": None,
        "num_threads": multiprocessing.cpu_count(),
        "dump": [],  # Supports: c, relay, tir, ll
        "disable_vectorize": "auto",
        "custom_unroll": False,  # Experimental, RISC-V only
        "autotuned_mode": None,
        "autotuned_results_file": None,
        "relay_debug": None,  # Use "DEFAULT=2" to have most verbosity. Needs USE_RELAY_DEBUG during setup.
        "refresh_model_info": False,
        "generate_wrapper": "auto",
    }

    REQUIRED = set()

    OPTIONAL = {"tvm.build_dir", "tvm.pythonpath", "tvm.configs_dir", "tvm.use_tlcpack"}

    def __init__(
        self, target="c", executor=None, runtime="crt", fmt="mlf", system_lib=False, features=None, config=None
    ):
        super().__init__(framework="tvm", features=features, config=config)

        self.model = None  # Actual filename!
        self.model_info = None
        self.input_shapes = None
        self.model_format = None
        self.supported_formats = get_supported_formats()
        self.target = target
        self.runtime = runtime
        self.executor = executor
        self.fmt = fmt
        self.system_lib = system_lib

        self.prefix = "default"
        self.artifacts = (
            []
        )  # TODO: either make sure that ony one model is processed at a time or move the artifacts to the methods
        self._tuning_records = {}
        results_file = self.config.get("autotuned_results_file", None)
        tuner_name = self.config.get("autotuned_mode", None)
        if results_file is not None:
            assert tuner_name is not None
            self._tuning_records[tuner_name] = results_file

    # On the long term, we might support multiple TUNE stages in a single run
    # (i.e. to allow autotvm+graphtuner to be separated)
    # Hence
    # @property
    # def tuning_records(self):
    #     if self._tuning_records:
    #         return self._tuning_records
    #     return self.config.get("autotuning_results_file", None):

    # @tuning_records.setter
    # def tuning_records(self, filepath):
    #     self._tuning_records = filepath

    def set_tuning_records(self, records, tuner_name=None):
        if tuner_name is None:
            tuner_name = self.config["autotuned_mode"]
            # tuner_name = "autotvm"
            assert tuner_name is not None
        self._tuning_records[tuner_name] = records

    def get_tuning_records(self, tuner_name=None):
        if tuner_name is None:
            tuner_name = self.config["autotuned_mode"]
            # tuner_name = "autotvm"
            if tuner_name is None:
                if len(self._tuning_records) > 0:
                    tuner_name = list(self._tuning_records.keys())[0]
        return self._tuning_records.get(tuner_name, None)

    @property
    def disable_vectorize(self):
        temp = self.config["disable_vectorize"]
        if temp is None or (isinstance(temp, str) and temp in ["auto", "AUTO"]):
            return self.target == "c"
        return str2bool(temp)

    @property
    def pass_config(self):
        base = {"tir.disable_vectorize": self.disable_vectorize}
        extra = self.config["extra_pass_config"]
        if isinstance(extra, str):
            import ast

            extra = ast.literal_eval(extra)
        assert isinstance(extra, dict)
        base.update(extra)
        return base

    @property
    def target_device(self):
        return self.config["target_device"]

    @property
    def target_mcpu(self):
        return self.config["target_mcpu"]

    @property
    def target_march(self):
        return self.config["target_march"]

    @property
    def target_mtriple(self):
        return self.config["target_mtriple"]

    @property
    def target_mabi(self):
        return self.config["target_mabi"]

    @property
    def target_mattr(self):
        return self.config["target_mattr"]

    @property
    def target_keys(self):
        return self.config["target_keys"]

    @property
    def target_model(self):
        return self.config["target_model"]

    @property
    def target_num_cores(self):
        return self.config["target_num_cores"]

    # TODO:
    # "target_device": ?,
    # "target_libs": ?,
    # "target_tag": ?,
    # "target_march": ?,
    # "target_keys": ?,
    # "target_opt_level": ?,
    # "target_cl_opt": ?,
    # "target_mfloat_abi": ?,
    # "target_fast_math_ninf": ?,
    # "target_fast_math_contract": ?,
    # "target_fast_math_nnan": ?,
    # "target_fast_math": ?,
    # "target_fast_math_nsz": ?,
    # "target_fast_math_reassoc": ?,
    # "target_fast_math_arcp": ?,

    @property
    def extra_targets(self):
        return str2list(self.config["extra_targets"], allow_none=True)

    @property
    def extra_target_details(self):
        return str2dict(self.config["extra_target_details"], allow_none=True)

    @property
    def desired_layout(self):
        return str2list(self.config["desired_layout"], allow_none=True)

    @property
    def desired_layout_ops(self):
        return str2list(self.config["desired_layout_ops"], allow_none=True)

    @property
    def desired_layout_map(self):
        return str2dict(self.config["desired_layout_map"], allow_none=True)

    @property
    def opt_level(self):
        return self.config["opt_level"]

    @property
    def use_tuning_results(self):
        value = self.config["use_tuning_results"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def tvmc_extra_args(self):
        return self.config["tvmc_extra_args"]

    @property
    def tvmc_custom_script(self):
        return self.config["tvmc_custom_script"]

    @property
    def disabled_passes(self):
        value = self.config["disabled_passes"]
        return str2list(value) if isinstance(value, str) else value

    @property
    def tvm_pythonpath(self):
        return self.config["tvm.pythonpath"]

    @property
    def tvm_build_dir(self):
        return self.config["tvm.build_dir"]

    @property
    def tvm_configs_dir(self):
        return self.config["tvm.configs_dir"]

    @property
    def tophub_url(self):
        return self.config["tophub_url"]

    @property
    def print_outputs(self):
        value = self.config["print_outputs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def use_tlcpack(self):
        value = self.config["tvm.use_tlcpack"]
        return str2bool(value, allow_none=True) if not isinstance(value, (bool, int)) else value

    @property
    def custom_unroll(self):
        value = self.config["custom_unroll"]
        return str2bool(value, allow_none=True) if not isinstance(value, (bool, int)) else value

    @property
    def dump(self):
        value = self.config["dump"]
        if isinstance(value, str):
            if "," in value:
                value = value.split(",")
            else:
                value = [value]
        for v in value:
            assert v in ["relay", "c", "ll", "tir"]
        assert isinstance(value, list)
        return value

    @property
    def needs_target(self):
        return self.target == "llvm"  # not c

    @property
    def refresh_model_info(self):
        value = self.config["refresh_model_info"]
        return str2bool(value, allow_none=True) if not isinstance(value, (bool, int)) else value

    @property
    def generate_wrapper(self):
        value = self.config["generate_wrapper"]
        if isinstance(value, str):
            if value.lower() == "auto":
                value = self.fmt == "mlf"
            else:
                value = str2bool(value)
        assert isinstance(value, bool)
        return value

    @property
    def num_threads(self):
        return self.config["num_threads"]

    @property
    def relay_debug(self):
        return self.config["relay_debug"]

    def get_target_details(self):
        ret = {}
        if self.target_device:
            ret["device"] = self.target_device
        if self.target_mcpu:
            ret["mcpu"] = self.target_mcpu
        if self.target_march:
            ret["march"] = self.target_march
        if self.target_mtriple:
            ret["mtriple"] = self.target_mtriple
        if self.target_mabi:
            ret["mabi"] = self.target_mabi
        if self.target_mattr:
            temp = self.target_mattr
            if self.custom_unroll:
                temp += ",+no-default-unroll"
            ret["mattr"] = temp
        if self.target_keys:
            ret["keys"] = self.target_keys
        if self.target_model:
            ret["model"] = self.target_model
        if self.target_num_cores:
            ret["num-cores"] = self.target_num_cores
        return ret

    def get_tvmc_compile_args(self, out, dump=None):
        assert self.executor is not None
        assert self.executor in ["aot", "graph"], "Unsupported TVM executor"
        args = [
            self.model,
            *get_target_tvmc_args(
                self.target,
                extra_targets=self.extra_targets,
                target_details=self.get_target_details(),
                extra_target_details=self.extra_target_details,
            ),
            *get_runtime_executor_tvmc_args(self.runtime, self.executor),
            *get_pass_config_tvmc_args(self.pass_config),
            *get_disabled_pass_tvmc_args(self.disabled_passes),
            *get_input_shapes_tvmc_args(self.input_shapes),
            *get_tuning_records_tvmc_args(self.use_tuning_results, self.get_tuning_records()),
            *get_desired_layout_args(self.desired_layout, self.desired_layout_ops, self.desired_layout_map),
            *(["--dump-code", ",".join(dump)] if dump is not None and len(dump) > 0 else []),
            *self.tvmc_extra_args,
            *["--opt-level", str(self.opt_level)],
            *["--output", str(out)],
            *["-f", self.fmt],
            *["--model-format", self.model_format],
        ]
        return args

    def invoke_tvmc(self, command, *args, cwd=None):
        env = prepare_python_environment(
            None if self.use_tlcpack else self.tvm_pythonpath,
            None if self.use_tlcpack else self.tvm_build_dir,
            None if self.use_tlcpack else self.tvm_configs_dir,
            tophub_url=self.tophub_url,
            num_threads=self.num_threads,
            debug_cfg=self.relay_debug,
        )
        if self.use_tlcpack:
            pre = ["tvmc"]
            return utils.exec_getout(*pre, command, *args, live=self.print_outputs, env=env, cwd=cwd)
        else:
            if self.tvmc_custom_script is None:
                pre = ["-m", "tvm.driver.tvmc"]
            else:
                pre = [self.tvmc_custom_script]
            return utils.python(*pre, command, *args, live=self.print_outputs, env=env, cwd=cwd)

    def invoke_tvmc_compile(self, out, dump=None, cwd=None):
        args = self.get_tvmc_compile_args(out, dump=dump)
        # self.timeout_sec = 90
        self.timeout_sec = 0
        if self.timeout_sec > 0:
            ret = exec_timeout(
                self.timeout_sec,
                self.invoke_tvmc,
                "compile",
                *args,
                cwd=cwd,
            )
        else:
            ret = self.invoke_tvmc("compile", *args, cwd=cwd)
        return ret

    def load_model(self, model, input_shapes=None, output_shapes=None, input_types=None, output_types=None):
        self.model = model
        # TODO: path model class instead of path!
        # fmt = self.model.formats[0]
        need_model_info = True
        if input_shapes:
            self.input_shapes = input_shapes
            if output_shapes and input_types and output_types:
                need_model_info = False
                self.model_format, self.model_info = get_fallback_model_info(
                    model, input_shapes, output_shapes, input_types, output_types, backend_name=self.name
                )
        if need_model_info:
            try:
                self.model_format, self.model_info = get_model_info(model, backend_name=self.name)
            except Exception as e:
                self.model_format = get_model_format(model)
                if self.model_format != "relay":
                    logger.warning(
                        "Fetching of Model Info failed (%s). Falling back to Relay-based info.", type(e).__name__
                    )
                    self.model_info = None
                else:
                    raise e

            if self.model_info:
                # TODO: also handle output_shapes
                # TODO: take care of refresh_model_info
                if input_shapes:
                    self.model_info.in_tensors = [t for t in self.model_info.in_tensors if t.name in self.input_shapes]
                else:
                    self.input_shapes = {tensor.name: tensor.shape for tensor in self.model_info.in_tensors}
        if self.model_info:
            self.model_info.validate()

    def get_graph_and_params_from_mlf(self, path):
        graph = None
        with open(Path(path) / "executor-config" / "graph" / "default.graph", "r") as handle:
            graph = handle.read()
        params = None
        with open(Path(path) / "parameters" / "default.params", "rb") as handle:
            params = handle.read()

        return graph, params

    def generate(self) -> Tuple[dict, dict]:
        artifacts = []
        assert self.model is not None
        dump = self.dump
        if self.refresh_model_info or (self.generate_wrapper and not self.model_info) and "relay" not in dump:
            dump.append("relay")
        with tempfile.TemporaryDirectory() as temp_dir:
            out_path = Path(temp_dir) / f"{self.prefix}.tar"
            out = self.invoke_tvmc_compile(out_path, dump=dump, cwd=temp_dir)
            if self.fmt == "mlf":
                mlf_path = Path(temp_dir) / "mlf"
                tarfile.open(out_path).extractall(mlf_path)
                with open(mlf_path / "metadata.json", "r") as handle:
                    metadata_txt = handle.read()
                artifacts.append(
                    Artifact(
                        f"{self.prefix}.json",
                        content=metadata_txt,
                        fmt=ArtifactFormat.TEXT,
                    )
                )
            with open(out_path, "rb") as handle:
                data = handle.read()
                artifacts.append(
                    Artifact(
                        f"{self.prefix}.tar",
                        raw=data,
                        fmt=ArtifactFormat.SHARED_OBJECT if self.fmt == "so" else ArtifactFormat.MLF,
                        archive=True,
                    )
                )
            if "c" in dump:
                with open(str(out_path) + ".c", "r") as handle:
                    mod_src = handle.read()
                    artifacts.append(
                        Artifact(
                            f"{self.prefix}.c",
                            content=mod_src,
                            fmt=ArtifactFormat.SOURCE,
                            optional=True,
                        )
                    )
            if "relay" in dump:
                with open(str(out_path) + ".relay", "r") as handle:
                    mod_txt = handle.read()
                    artifacts.append(
                        Artifact(
                            f"{self.prefix}.relay",
                            content=mod_txt,
                            fmt=ArtifactFormat.TEXT,
                            optional=True,
                        )
                    )
            if "ll" in dump:
                with open(str(out_path) + ".ll", "r") as handle:
                    mod_txt = handle.read()
                    artifacts.append(
                        Artifact(
                            f"{self.prefix}.ll",
                            content=mod_txt,
                            fmt=ArtifactFormat.SOURCE,
                            optional=True,
                        )
                    )
            if self.executor == "graph":
                if self.fmt == "so":
                    pass
                    # raise NotImplementedError
                elif self.fmt == "mlf":
                    graph, params = self.get_graph_and_params_from_mlf(mlf_path)
                    artifacts.append(
                        Artifact(
                            f"{self.prefix}.graph",
                            content=graph,
                            fmt=ArtifactFormat.SOURCE,
                        )
                    )
                    artifacts.append(
                        Artifact(
                            f"{self.prefix}.params",
                            raw=params,
                            fmt=ArtifactFormat.RAW,
                        )
                    )
                else:
                    raise RuntimeError("Unsupported fmt")
        stdout_artifact = Artifact(
            "tvmc_compile_out.log", content=out, fmt=ArtifactFormat.TEXT
        )  # TODO: rename to tvmaot_out.log?
        artifacts.append(stdout_artifact)
        return {"default": artifacts}, {"default": Metrics()}



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TVM Tune Platform"""
import re
import os
from mlonmcu.config import str2bool
import time
import tempfile
import tarfile
import concurrent
from pathlib import Path
from .tvm_target_platform import TvmTargetPlatform
from ..platform import TunePlatform

from mlonmcu.flow.tvm.backend.tuner import (
    get_autotuning_defaults,
    get_autotvm_defaults,
    get_autoscheduler_defaults,
    get_metascheduler_defaults,
)
from mlonmcu.flow.tvm.backend.tvmc_utils import (
    get_rpc_tvmc_args,
    get_target_tvmc_args,
    get_disabled_pass_tvmc_args,
)
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.target.metrics import Metrics
from mlonmcu.flow.tvm.backend.python_utils import prepare_python_environment
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

logger = get_logger()


class TvmTunePlatform(TunePlatform, TvmTargetPlatform):
    """TVM Tune platform class."""

    FEATURES = TunePlatform.FEATURES | TvmTargetPlatform.FEATURES | {"autotvm", "autoscheduler", "metascheduler"}

    DEFAULTS = {
        **TunePlatform.DEFAULTS,
        **TvmTargetPlatform.DEFAULTS,
        "experimental_tvmc_tune_tasks": False,
        "experimental_tvmc_tune_visualize": False,
        # "experimental_tvmc_tune_wandb": False,  # TODO
        "enable_wandb": False,
        "min_repeat_ms": 0,
        **{("autotuning_" + key): value for key, value in get_autotuning_defaults().items()},
        **{("autotvm_" + key): value for key, value in get_autotvm_defaults().items()},
        **{("autoscheduler_" + key): value for key, value in get_autoscheduler_defaults().items()},
        **{("metascheduler_" + key): value for key, value in get_metascheduler_defaults().items()},
    }

    REQUIRED = TunePlatform.REQUIRED | TvmTargetPlatform.REQUIRED

    @property
    def tune_tasks(self):
        # Effectively select which tasks should be tuned in the session
        return self.config["tune_tasks"]

    @property
    def experimental_tvmc_tune_tasks(self):
        value = self.config["experimental_tvmc_tune_tasks"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def experimental_tvmc_tune_visualize(self):
        value = self.config["experimental_tvmc_tune_visualize"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_wandb(self):
        value = self.config["enable_wandb"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def min_repeat_ms(self):
        value = self.config["min_repeat_ms"]
        return int(value)

    def invoke_tvmc_tune(self, *args, target=None, **kwargs):
        return self.invoke_tvmc("tune", *args, target=target, **kwargs)

    def get_tune_args(self, model, backend, target, out, trials, early_stopping):
        max_parallel = int(self.config.get("autotuning_max_parallel", 1))
        timeout = int(self.config.get("autotuning_timeout", 1000))
        results_file = self.config.get("autotuning_results_file", None)
        desired_layout = backend.config.get("desired_layout", None)
        ret = [
            *get_target_tvmc_args(
                backend.target,
                extra_targets=backend.extra_targets,
                target_details=backend.get_target_details(),
                extra_target_details=backend.extra_target_details,
            ),
            *(["--desired-layout", desired_layout] if desired_layout is not None else []),
            *get_rpc_tvmc_args(self.use_rpc, self.rpc_key, self.rpc_hostname, self.rpc_port),
            *get_disabled_pass_tvmc_args(backend.disabled_passes),
            # TODO: missing: pass config etc.
            *(["--early-stopping", str(early_stopping)] if early_stopping > 0 else []),
            *["--parallel", str(max_parallel)],
            *["--timeout", str(timeout * max_parallel)],
            *["--trials", str(trials)],
            *["--number", str(self.number)],  # TODO: increase while tuning?
            *["--repeat", str(self.repeat)],  # TODO: increase while tuning?
            *["--min-repeat-ms", str(self.min_repeat_ms)],
            *(["--tuning-records", results_file] if results_file is not None else []),
            *["--output", str(out)],
        ]
        if self.config["autotuning_tasks"]:
            assert self.experimental_tvmc_tune_tasks, f"{self.name}.tune_tasks requires experimental_tvmc_tune_tasks"
            ret.extend(["--tasks", str(self.config["autotuning_tasks"])])
        if self.enable_wandb:
            ret.append("--wandb-callback")
        ret.append(model)
        return ret

    def get_autotvm_tune_args(self, model, backend, target, out, trials_global, early_stopping):
        ret = self.get_tune_args(model, backend, target, out, trials_global, early_stopping)

        tuner = self.config.get("autotvm_tuner", "ga")
        assert tuner in ["ga", "gridsearch", "random", "xgb", "xgb_knob", "xgb-rank"]
        ret.extend(["--tuner", tuner])
        if self.config["autotuning_visualize"]:
            to_file = self.config["autotuning_visualize_file"]
            if not to_file:
                to_file = "viz.png"
            live = self.config["autotuning_visualize_live"]
            assert self.experimental_tvmc_tune_tasks, "requires experimental_autotvm_visualize"
            visualize_arg = to_file
            if live:
                visualize_arg += ",live"
            ret.extend(["--visualize", visualize_arg])
        return ret

    def get_autoscheduler_tune_args(self, model, backend, target, out, trials_global, early_stopping):
        assert not self.enable_wandb, "WANDB callback not yet supported by AutoScheduler"
        ret = self.get_tune_args(model, backend, target, out, trials_global, early_stopping)
        ret.append("--enable-autoscheduler")
        if self.config.get("autoscheduler_include_simple_tasks", False):
            ret.append("--include-simple-tasks")
        if self.config.get("autoscheduler_log_estimated_latency", False):
            ret.append("--log-estimated-latency")
        hardware_details = target.get_hardware_details()
        if len(hardware_details) > 0:
            for key, value in hardware_details.items():
                ret.extend([f"--{key}", str(value)])
        return ret

    def get_metascheduler_tune_args(self, model, backend, target, out, trials_global, trials_single, early_stopping):
        ret = self.get_tune_args(model, backend, target, out, trials_global, early_stopping)
        ret.append("--enable-metascheduler")
        if trials_single:
            ret.extend(["--trials-per-task", str(trials_single)])
        return ret

    def _tune_model(self, model_path, backend, target):
        autotvm_enable = self.config["autotvm_enable"]
        autoscheduler_enable = self.config["autoscheduler_enable"]
        metascheduler_enable = self.config["metascheduler_enable"]
        if not autotvm_enable and not autoscheduler_enable and not metascheduler_enable:
            # Tuning not enabled! (Might happen if abstract autotune feature is used!)
            return {}, {}
        assert [autotvm_enable, autoscheduler_enable, metascheduler_enable].count(
            True
        ) == 1, "Can not use AutoTVM and AutoScheduler/MetaScheduler at the same time"
        results_file = self.config["autotuning_results_file"]
        append = self.config["autotuning_append"]
        num_workers = self.config["autotuning_num_workers"]
        artifacts = []
        verbose = False
        if self.print_outputs:
            verbose = True

        def remove_empty(inp):
            return [line for line in inp if len(line.strip()) > 0]

        def count_failed_trials(inp):
            cnt = 0
            for line in inp.split("\n"):
                m = re.compile(r".*\[1000000000\.0\].*").match(line)
                if m:
                    cnt += 1
            return cnt

        def get_max_flops(out, prefix="M"):
            res = re.compile(r"\d+\.\d+\s*\/\s*(\d+\.\d+)\s+{prefix}FLOPS").findall(out)
            if len(res) > 0:
                return res[-1]
            return -1

        # pick best records
        def _pick_best(backend, records, verbose=False):
            with tempfile.TemporaryDirectory() as tmp_dir:
                in_file = Path(tmp_dir) / "tuning_results.log.txt"
                with open(in_file, "w") as handle:
                    handle.write(records)
                out_file = Path(tmp_dir) / "best_tuning_results.log.txt"
                args = [
                    "--mode",
                    "pick",
                    "--i",
                    in_file,
                    "--o",
                    out_file,
                ]
                env = prepare_python_environment(backend.tvm_pythonpath, backend.tvm_build_dir, backend.tvm_configs_dir)
                utils.python("-m", "tvm.autotvm.record", *args, live=verbose, env=env)
                with open(out_file, "r") as handle:
                    content_best = handle.read()
            return content_best

        content = ""
        total_size = None
        visualize_raw = None
        if num_workers is not None:
            if isinstance(num_workers, str):
                num_workers = int(num_workers)
            assert isinstance(num_workers, int) and num_workers >= 0
        trials_global = self.config.get("autotuning_trials", 10)
        trials_single = self.config.get("autotuning_trials_single", None)
        if not isinstance(trials_global, int):
            trials_global = int(trials_global)
        if trials_single is not None and not isinstance(trials_single, int):
            trials_single = int(trials_single)
        trials = trials_single if trials_single is not None else trials_global
        assert isinstance(trials, int)
        early_stopping = self.config.get("autotuning_early_stopping", None)
        if early_stopping is None:
            early_stopping = max(trials, 10)  # Let's see if this default works out...
        if not isinstance(early_stopping, int):
            early_stopping = int(early_stopping)
        assert isinstance(early_stopping, int)
        if metascheduler_enable:
            assert not append, "append not supported by MetaScheduler"
            assert num_workers is None or int(num_workers) == 0, "num_workers > 0 not supported by MetaScheduler"
            assert not self.config["autotuning_visualize"], "autotuning_visualize not supported by MetaScheduler"

            sub_metrics = {}
            sub_artifacts = {}
            with tempfile.TemporaryDirectory() as tmp_dir:
                # out_file = Path(tmp_dir) / "tuning_results.log.txt"
                tmp_dir = Path(tmp_dir)
                work_dir = tmp_dir / "work_dir"
                tune_args = self.get_metascheduler_tune_args(
                    model_path, backend, target, work_dir, trials_global, trials_single, 0
                )
                out = self.invoke_tvmc_tune(*tune_args, target=target, cwd=tmp_dir)
                with tarfile.open(tmp_dir / "work_dir.tar", "w") as tar:
                    for file in os.listdir(work_dir):
                        tar.add(work_dir / file, arcname=os.path.join("work_dir", file))
                raw = None
                with open(tmp_dir / "work_dir.tar", "rb") as tar:
                    raw = tar.read()
                artifact = Artifact(
                    "work_dir.tar", raw=raw, fmt=ArtifactFormat.ARCHIVE, flags=["records", "metascheduler"]
                )
        elif autotvm_enable or autoscheduler_enable:
            if append:
                if results_file is not None:
                    with open(results_file, "r") as handle:
                        content = handle.read()

            sub_metrics = {}
            sub_artifacts = {}
            if num_workers is not None and num_workers > 0:
                assert self.experimental_tvmc_tune_tasks, "num_workers requires experimental_tvmc_tune_tasks=1"
                # TODO: fix
                assert self.config["autotuning_tasks"] is None, "tune_tasks not supported together with num_workers > 0"

                def get_tune_tasks():
                    with tempfile.TemporaryDirectory() as tmp_dir:
                        out_file = Path(tmp_dir) / "tuning_results.log.txt"
                        if autotvm_enable:
                            tune_args = self.get_autotvm_tune_args(model_path, backend, target, out_file, 1, 0)
                        elif autoscheduler_enable:
                            tune_args = self.get_autoscheduler_tune_args(model_path, backend, target, out_file, 1, 0)
                        else:
                            assert False
                        out = self.invoke_tvmc_tune(
                            *tune_args, "--tasks", "list", target=target, cwd=tmp_dir, live=False
                        )
                        lines = out.split("\n")
                        for i, line in enumerate(lines):
                            if "Available Tasks for tuning" in line:
                                lines = lines[i + 1 :]
                                break
                        # tasks = [line.split(". ", 1)[1] for line in lines if len(line.strip()) > 0]
                        # Get config space sizes
                        matches = re.compile(r"(\d+). Task.*\(len=(\d+)\)").findall(out)
                        sizes = list(map(lambda x: (int(x[0]), int(x[1])), matches))
                        return sizes

                # num_tasks = len(get_tune_tasks())
                tune_tasks = get_tune_tasks()
                workers = []
                with concurrent.futures.ThreadPoolExecutor(num_workers) as executor:
                    # for i in range(num_tasks):
                    for i, task_len in tune_tasks:
                        if total_size is None:
                            total_size = 0
                        total_size += task_len
                        logger.debug(f"Created worker for task {i}")

                        def do_work(idx, prepend, task_len):
                            nonlocal trials_single, trials_global, early_stopping
                            t0 = time.time()
                            with tempfile.TemporaryDirectory() as tmp_dir:
                                out_file = Path(tmp_dir) / "tuning_results.log.txt"
                                with open(out_file, "w") as handle:
                                    handle.write(prepend)
                                if trials_single == 0 or (
                                    trials_single is None
                                ):  # 0: auto, None: do not limit per task
                                    trials_single = max(1, trials_global // len(tune_tasks))
                                    early_stopping = max(trials_single, 10)  # Let's see if this default works out...
                                if autotvm_enable:
                                    tune_args = self.get_autotvm_tune_args(
                                        model_path, backend, target, out_file, trials_single, early_stopping
                                    )
                                elif autoscheduler_enable:
                                    tune_args = self.get_autoscheduler_tune_args(
                                        model_path, backend, target, out_file, trials_single, early_stopping
                                    )
                                else:
                                    assert False
                                out = self.invoke_tvmc_tune(*tune_args, "--tasks", str(idx), target=target, cwd=tmp_dir)
                                with open(out_file, "r") as handle:
                                    content = handle.read()
                                visualize_raw_task = None
                                if self.config["autotuning_visualize"]:
                                    to_file = self.config["autotuning_visualize_file"]
                                    if not to_file or to_file is True:
                                        to_file = Path(tmp_dir) / "viz.png"
                                    else:
                                        to_file = Path(to_file)
                                    assert to_file.is_file()
                                    with open(to_file, "rb") as handle:
                                        visualize_raw_task = handle.read()
                                # content_best = _pick_best(backend, content, verbose=verbose)
                                sub_trials = len(remove_empty(content.split("\n")))
                                sub_failed_trials = count_failed_trials(content)
                                max_flops = get_max_flops(out)
                                t1 = time.time()
                            return (
                                out,
                                content,
                                task_len,
                                sub_trials,
                                sub_failed_trials,
                                max_flops,
                                t1 - t0,
                                visualize_raw_task,
                            )

                        workers.append(executor.submit(do_work, i, content, task_len))
                all_out = ""
                all_content = ""
                for i, w in enumerate(workers):
                    logger.debug(f"Worker {i}: pending")
                    metrics_ = Metrics()
                    artifacts_ = []
                    try:
                        ret = w.result()
                        logger.debug(f"Worker {i}: done")
                        out, content, size, tuned, failed, max_flops, duration, visualize_raw_task = ret
                        all_out += out
                        all_content += content
                        metrics_.add("Config Space Size", size, True)
                        metrics_.add("Total Trials", tuned, True)
                        metrics_.add("Failed Trials", failed, True)
                        metrics_.add("Max. MFLOPS", max_flops, True)
                        metrics_.add("Tune Duration [s]", duration, True)
                        metrics_.add("Tune Duration per Trial [s]", duration / tuned + failed, True)
                        if early_stopping < trials_single:
                            early = tuned + failed < min(trials_single, size)
                        else:
                            early = False
                        metrics_.add("Early Stopped", early, True)
                        if visualize_raw_task:
                            visualize_artifact = Artifact(
                                f"tuning_progress_task{i}.png",
                                raw=visualize_raw_task,
                                fmt=ArtifactFormat.RAW,
                                flags=["visualize"],
                            )
                            artifacts_.append(visualize_artifact)
                    except AssertionError:
                        logger.exception(f"Worker {i}: failed")
                        metrics_.add("Failed Tuning", True)
                    sub_metrics[f"task{i}"] = metrics_
                    sub_artifacts[f"task{i}"] = artifacts_
                out = all_out
                content = all_content
            else:
                with tempfile.TemporaryDirectory() as tmp_dir:
                    out_file = Path(tmp_dir) / "tuning_results.log.txt"
                    with open(out_file, "w") as handle:
                        handle.write(content)
                    if autotvm_enable:
                        tune_args = self.get_autotvm_tune_args(
                            model_path, backend, target, out_file, trials_global, early_stopping
                        )
                    elif autoscheduler_enable:
                        tune_args = self.get_autoscheduler_tune_args(
                            model_path, backend, target, out_file, trials_global, early_stopping
                        )  # TODO: expose per_task trials
                    else:
                        assert False
                    out = self.invoke_tvmc_tune(*tune_args, target=target, cwd=tmp_dir)
                    with open(out_file, "r") as handle:
                        content = handle.read()
                    visualize_raw = None
                    if self.config["autotuning_visualize"]:
                        to_file = self.config["autotuning_visualize_file"]
                        if not to_file or to_file is True:
                            to_file = Path(tmp_dir) / "viz.png"
                        else:
                            to_file = Path(tmp_dir)
                        assert to_file.is_file()
                        with open(to_file, "rb") as handle:
                            visualize_raw = handle.read()
        else:
            if results_file is None:
                return {}, {}
            assert Path(results_file).is_file()
            with open(results_file, "r") as handle:
                content = handle.read()

        if metascheduler_enable:
            artifacts.append(artifact)
            # TODO: get num trials etc.
            metrics = Metrics()
        elif autotvm_enable or autoscheduler_enable:
            flag = "autotvm" if not autoscheduler_enable else "autoscheduler"
            artifact = Artifact(
                "tuning_results.log.txt", content=content, fmt=ArtifactFormat.TEXT, flags=["records", flag]
            )
            artifacts.append(artifact)
            if visualize_raw:
                visualize_artifact = Artifact(
                    "tuning_progress.png", raw=visualize_raw, fmt=ArtifactFormat.RAW, flags=["visualize"]
                )
                artifacts.append(visualize_artifact)

            metrics = Metrics()

            if total_size is not None:
                metrics.add("Config Space Size", total_size, True)

            content_best = _pick_best(backend, content, verbose=verbose)
            total_trials = len(remove_empty(content.split("\n")))
            metrics.add("Total Trials", total_trials, True)

            failed_trials = count_failed_trials(content)
            metrics.add("Failed Trials", failed_trials, True)
            if len(content_best) > 0:
                artifact_ = Artifact("best_tuning_results.log.txt", content=content_best, fmt=ArtifactFormat.TEXT)
                artifacts.append(artifact_)
                num_tuned = len(remove_empty(content_best.split("\n")))
                metrics.add("Tuned Tasks", num_tuned, True)
            else:
                artifact_ = Artifact("best_tuning_results.log.txt", content="", fmt=ArtifactFormat.TEXT)
                artifacts.append(artifact_)
                metrics.add("Tuned Tasks", 0, True)

        if autotvm_enable or autoscheduler_enable or metascheduler_enable:
            stdout_artifact = Artifact(
                "tvmc_tune_out.log", content=out, fmt=ArtifactFormat.TEXT
            )  # TODO: rename to tvmaot_out.log?
            artifacts.append(stdout_artifact)

        return {"default": artifacts, **sub_artifacts}, {"default": metrics, **sub_metrics}



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Host/x86 Target definitions"""

import stat
from pathlib import Path

from mlonmcu.config import str2bool
from mlonmcu.setup.utils import execute
from .common import cli
from .target import Target


class HostX86Target(Target):
    """Target using the x86 host system

    Mainly interesting to easy testing and debugging because benchmarking is not possible.
    """

    FEATURES = Target.FEATURES | {"gdbserver"}

    DEFAULTS = {
        **Target.DEFAULTS,
        "gdbserver_enable": False,
        "gdbserver_attach": False,
        "gdbserver_port": 2222,
    }

    def __init__(self, name="host_x86", features=None, config=None):
        super().__init__(name, features=features, config=config)
        self.gdb_path = "gdb"
        self.gdb_server_path = "gdbserver"

    @property
    def gdbserver_enable(self):
        value = self.config["gdbserver_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_attach(self):
        value = self.config["gdbserver_attach"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_port(self):
        return int(self.config["gdbserver_port"])

    def exec(self, program, *args, **kwargs):
        def make_executable(exe):
            f = Path(exe)
            f.chmod(f.stat().st_mode | stat.S_IEXEC)

        make_executable(program)
        if self.gdbserver_enable:
            if self.gdbserver_attach:
                raise NotImplementedError
                # return execute(self.gdb_path, program, *args, **kwargs)
            else:
                comm = f"127.0.0.1:{self.gdbserver_port}"
                return execute(self.gdb_server_path, comm, program, *args, **kwargs)

        return execute(program, *args, **kwargs), []

    def get_arch(self):
        return "x86"


if __name__ == "__main__":
    cli(target=HostX86Target)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU ARA Target definitions"""

import os
import re
from pathlib import Path
from tempfile import TemporaryDirectory

# import time
import multiprocessing

from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool

# from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.bench import add_bench_metrics
from .riscv_vext_target import RVVTarget
from .util import update_extensions

logger = get_logger()


class AraTarget(RVVTarget):
    """Target using a Pulpino-like VP running in the GVSOC simulator"""

    FEATURES = RVVTarget.FEATURES | {"log_instrs", "vext"}

    DEFAULTS = {
        **RVVTarget.DEFAULTS,
        "xlen": 64,
        "nr_lanes": 4,
        "vlen": 4096,  # default value for hardware compilation, will be overwritten by -c vext.vlen
        "enable_vext": False,
        "vext_spec": 1.0,
        "embedded_vext": False,
        "elen": 64,
        "num_threads": multiprocessing.cpu_count(),
        "limit_cycles": 10000000,
    }

    REQUIRED = RVVTarget.REQUIRED | {
        "ara.src_dir",  # for the bsp package
        "verilator.install_dir",  # for simulation
    }

    OPTIONAL = RVVTarget.OPTIONAL | {
        "ara.verilator_tb",
    }

    def __init__(self, name="ara", features=None, config=None):
        super().__init__(name, features=features, config=config)
        assert self.config["xlen"] == 64, 'ARA target must has xlen equal 64, try "-c ara.xlen=64"'

    @property
    def ara_apps_dir(self):
        return Path(self.config["ara.src_dir"]) / "apps"

    @property
    def ara_hardware_dir(self):
        return Path(self.config["ara.src_dir"]) / "hardware"

    @property
    def verilator_install_dir(self):
        return Path(self.config["verilator.install_dir"])

    @property
    def ara_verilator_tb(self):
        value = self.config.get("ara.verilator_tb", None)
        return Path(value) if value is not None else None

    @property
    def nr_lanes(self):
        value = self.config["nr_lanes"]
        return value

    @property
    def vlen(self):
        value = self.config["vlen"]
        return value

    @property
    def limit_cycles(self):
        value = self.config["limit_cycles"]
        return int(value) if value is not None else None

    @property
    def enable_vext(self):
        value = self.config["enable_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def vext_spec(self):
        return float(self.config["vext_spec"])

    @property
    def embedded_vext(self):
        value = self.config["embedded_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def elen(self):
        return int(self.config["elen"])

    @property
    def extensions(self):
        exts = super().extensions
        return update_extensions(
            exts,
        )

    @property
    def num_threads(self):
        return self.config["num_threads"]

    def prepare_simulator(self, program, *_, cwd=os.getcwd(), **kwargs):
        # populate the ara verilator testbench directory
        self.tb_ara_verilator_build_dir = TemporaryDirectory()
        # self.tb_ara_verilator_build_dir = Path("/tmp/tmpd7kz4i6z")
        # env = os.environ.copy()
        # env["ROOT_DIR"] = str(self.ara_hardware_dir)
        # env["veril_library"] = self.tb_ara_verilator_build_dir.name
        # env["veril_path"] = str(self.verilator_install_dir / "bin")
        # env["nr_lanes"] = str(self.nr_lanes)
        # env["vlen"] = str(self.vlen)
        # env["bender_defs"] = f"--define NR_LANES={self.nr_lanes} --define VLEN={self.vlen} --define RVV_ARIANE=1"
        args = []
        args.append(f"ROOT_DIR={self.ara_hardware_dir}")
        args.append(f"veril_library={self.tb_ara_verilator_build_dir.name}")
        args.append(f"veril_path={self.verilator_install_dir}/bin")
        args.append(f"nr_lanes={self.nr_lanes}")
        args.append(f"vlen={self.vlen}")
        compile_verilator_tb_ret = execute(
            "make",
            "verilate",
            f"JOBS={self.num_threads}",
            # env=env,
            cwd=self.ara_hardware_dir,
            *args,
            **kwargs,
        )
        return compile_verilator_tb_ret

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute an executable with given arguments"""
        # run simulation
        # to add trace: https://github.com/pulp-platform/ara/blob/main/hardware/Makefile#L201
        ara_verilator_args = []
        limit_cycles = self.limit_cycles
        if limit_cycles is not None:
            ara_verilator_args.extend(["-c", str(limit_cycles)])
        ara_verilator_args.extend(["-l", f"ram,{program}"])
        assert len(self.extra_args) == 0

        if self.ara_verilator_tb:
            exe = self.ara_verilator_tb
        else:
            assert (
                self.tb_ara_verilator_build_dir is not None
            ), "A folder containing Vara_tb_verilator should be generated by the function prepare_simulator"
            exe = Path(self.tb_ara_verilator_build_dir.name) / "Vara_tb_verilator"
        simulation_ret = execute(
            str(exe),
            *ara_verilator_args,
            # env=env,
            cwd=cwd,
            *args,
            **kwargs,
        )
        if not self.ara_verilator_tb:
            self.tb_ara_verilator_build_dir.cleanup()
        return simulation_ret

    def parse_exit(self, out):
        exit_code = super().parse_exit(out)
        if exit_code is None:
            if "Simulation timeout of" in out:
                exit_code = -1
        return exit_code

    def parse_stdout(self, out, metrics, exit_code=0):
        add_bench_metrics(out, metrics, exit_code != 0, target_name=self.name)
        """
        Expected output looks like this:

        Simulation statistics
        =====================
        Executed cycles:  5e437
        Wallclock time:   150.173 s
        Simulation speed: 2571.05 cycles/s (2.57105 kHz)
        """
        sim_insns = re.search(r"Executed cycles:\s+([0-9a-f]+)", out)
        if sim_insns:
            sim_insns = int(sim_insns.group(1), 16)
            metrics.add("Simulated Instructions", sim_insns, True)
        wall = re.search(r"Wallclock time:\s+(\d*.?\d*)\ss", out)
        if wall:
            wall = float(wall.group(1))
            metrics.add("Wallclock time", wall, True)
        speed = re.search(r"Simulation speed: (\d*\.?\d*) cycles\/s", out)
        if speed:
            speed = int(float(speed.group(1)))
            metrics.add("MIPS", speed / 1e6, True)

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""
        if not self.ara_verilator_tb:
            if self.print_outputs:
                self.prepare_simulator(elf, *args, cwd=directory, live=True, handle_exit=handle_exit)
            else:
                self.prepare_simulator(
                    elf,
                    *args,
                    cwd=directory,
                    live=False,
                    print_func=lambda *args, **kwargs: None,
                    handle_exit=handle_exit,
                )

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            # TODO: before or after?
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        # simulation_start = time.time()
        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
        # simulation_end = time.time()
        exit_code = 0
        metrics = Metrics()
        self.parse_stdout(out, metrics, exit_code=exit_code)
        return metrics, out, []

    def get_target_system(self):
        return self.name

    def get_platform_defs(self, platform):
        assert platform == "mlif"
        ret = super().get_platform_defs(platform)
        # ret["RISCV_ARCH"] = "rv32imcxpulpv3"
        ret["XLEN"] = self.xlen
        ret["RISCV_ABI"] = self.abi
        ret["ARA_APPS_DIR"] = self.ara_apps_dir
        ret["MLONMCU_ARA_NR_LANES"] = self.nr_lanes
        ret["MLONMCU_ARA_VLEN"] = self.vlen
        ret["CMAKE_VERBOSE_MAKEFILE"] = "BOOL=OFF"
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        return ret


if __name__ == "__main__":
    cli(target=AraTarget)



import pytest

from mlonmcu.target.riscv.riscv import RISCVTarget
from mlonmcu.target.riscv.riscv_vext_target import RVVTarget
from mlonmcu.target.riscv.riscv_pext_target import RVPTarget
from mlonmcu.target import register_target
from mlonmcu.environment.config import TargetConfig

from .helpers import MODEL_FRONTENDS, TARGET_PLATFORMS, _check_features, _init_run


def _check_multilib(context, gcc):
    environment = context.environment
    user_vars = environment.vars
    # TODO: properly lookup in dependency cache?
    return user_vars.get(f"{gcc}.multilib", False)


def _get_target_config(target, xlen, fpu, embedded, compressed, atomic, multiply):
    return {
        f"{target}.xlen": xlen,
        f"{target}.embedded": embedded,
        f"{target}.compressed": compressed,
        f"{target}.atomic": atomic,
        f"{target}.multiply": multiply,
        f"{target}.fpu": fpu,
    }


def _get_platform_config(platform, toolchain):
    return {
        f"{platform}.print_outputs": True,
        f"{platform}.toolchain": toolchain,
    }


def _get_feature_vext_config(vlen, elen, spec, embedded):
    return {
        "vext.vlen": vlen,
        "vext.elen": elen,
        "vext.spec": spec,
        "vext.embedded": embedded,
    }


def _get_feature_pext_config(spec):
    return {
        "pext.spec": spec,
    }


class MyRISCVTarget(RISCVTarget):
    """Base RISC-V target with support as we do not run simulations, just compile."""

    def __init__(self, name="myriscv_default", features=None, config=None):
        super().__init__(name, features=features, config=config)


class MyRVVTarget(RVVTarget):
    """RISC-V target with vector (rvv) support as we do not run simulations, just compile."""

    def __init__(self, name="myriscv_vector", features=None, config=None):
        super().__init__(name, features=features, config=config)


class MyRVPTarget(RVPTarget):
    """RISC-V target with packed (rvp) support as we do not run simulations, just compile."""

    def __init__(self, name="myriscv_packed", features=None, config=None):
        super().__init__(name, features=features, config=config)


register_target("myriscv_default", MyRISCVTarget)
register_target("myriscv_vector", MyRVVTarget)
register_target("myriscv_packed", MyRVPTarget)


def _test_compile_platform2(
    platform_name, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
):
    user_config = user_context.environment.vars.copy()
    user_config.update(config)
    frontend_name = MODEL_FRONTENDS[model_name]
    if platform_name is None:
        platform_name = TARGET_PLATFORMS[target_name]
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    if not user_context.environment.has_backend(backend_name):
        pytest.skip(f"Backend '{backend_name}' is not enabled.")
    if not user_context.environment.has_platform(platform_name):
        pytest.skip(f"Platform '{platform_name}' is not enabled.")  # TODO: not enabled -> not installed
    if not user_context.environment.has_target(target_name):
        pytest.skip(f"Target '{target_name}' is not enabled.")  # TODO: remove check?
    _check_features(user_context, feature_names)
    session, run = _init_run(user_context, models_dir, user_config)
    with session:
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        run.add_backend_by_name(backend_name, context=user_context)  # TODO: implicit Framework
        run.add_platform_by_name(platform_name, context=user_context)
        run.add_target_by_name(target_name, context=user_context)
        # arch = run.target.arch
        # abi = run.target.abi
        # vext = "vext" in features
        # pext = "pext" in features
        # toolchain = run.platform.toolchain
        # _check_available_multilibs(user_context, arch, abi,... , toolchain)
        # assert session.process_runs(until=RunStage.COMPILE, context=user_context)
    report = session.get_reports()
    df, artifacts = report.df, run.artifacts

    assert len(df) == 1
    assert df["Model"][0] == model_name
    assert df["Platform"][0] == platform_name
    assert df["Target"][0] == target_name
    return df, artifacts


def _test_riscv_toolchain(
    platform_name, backend_name, target_name, user_context, model_name, models_dir, feature_names, config, toolchain
):
    user_context.environment.targets.extend(
        [
            TargetConfig("myriscv_default"),
            TargetConfig("myriscv_vector"),
            TargetConfig("myriscv_packed"),
        ]
    )
    assert platform_name == "mlif"
    user_config = user_context.environment.vars.copy()
    user_config.update(_get_platform_config(platform_name, toolchain))
    if not user_context.environment.has_toolchain(toolchain):
        pytest.skip(f"Toolchain '{toolchain}' not available in user environment!")

    return _test_compile_platform2(
        platform_name, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from mlonmcu.flow.backend import Backend
from mlonmcu.models.model import ModelFormats


class TFLMBackend(Backend):
    registry = {}

    name = None

    FEATURES = set()

    DEFAULTS = {}

    REQUIRED = set()

    def __init__(self, features=None, config=None):
        super().__init__(framework="tflm", config=config, features=features)
        self.model = None
        self.supported_formats = [ModelFormats.TFLITE]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        assert isinstance(cls.name, str)
        cls.registry[cls.name] = cls

    def load_model(self, model, input_shapes=None, output_shapes=None, input_types=None, output_types=None):
        self.model = model
        ext = os.path.splitext(model)[1][1:]
        fmt = ModelFormats.from_extension(ext)
        assert fmt == ModelFormats.TFLITE, f"Backend '{self.name}' does not support model format: {fmt.name}"



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import yaml
import pathlib
import logging

from .config import PathConfig


def create_environment_dict(environment):
    data = {}
    data["home"] = environment.home
    data["logging"] = {
        "level": logging.getLevelName(environment.defaults.log_level),
        "to_file": environment.defaults.log_to_file,
        "rotate": environment.defaults.log_rotate,
    }
    data["cleanup"] = {
        "auto": environment.defaults.cleanup_auto,
        "keep": environment.defaults.cleanup_keep,
    }
    data["paths"] = {
        path: (
            str(path_config.path)
            if isinstance(path_config, PathConfig)
            else [str(config.path) for config in path_config]
        )
        for path, path_config in environment.paths.items()
    }  # TODO: allow relative paths
    data["repos"] = {repo: vars(repo_config) for repo, repo_config in environment.repos.items()}
    # TODO: test with options!
    data["frameworks"] = {
        "default": environment.defaults.default_framework if environment.defaults.default_framework else None,
        **{
            framework.name: {
                "enabled": framework.enabled,
                "backends": {
                    "default": (
                        environment.defaults.default_backends[framework.name]
                        if environment.defaults.default_backends
                        and environment.defaults.default_backends[framework.name]
                        else None
                    ),
                    **{
                        backend.name: {
                            "enabled": backend.enabled,
                            "features": {
                                backend_feature.name: backend_feature.supported for backend_feature in backend.features
                            },
                        }
                        for backend in framework.backends
                    },
                },
                "features": {
                    framework_feature.name: framework_feature.supported for framework_feature in framework.features
                },
            }
            for framework in environment.frameworks
        },
    }
    data["frontends"] = {
        # "default": None,  # unimplemented?
        **{
            frontend.name: {
                "enabled": frontend.enabled,
                "features": {
                    frontend_feature.name: frontend_feature.supported for frontend_feature in frontend.features
                },
            }
            for frontend in environment.frontends
        },
    }
    data["platforms"] = {
        # "default": None,  # unimplemented?
        **{
            platform.name: {
                "enabled": platform.enabled,
                "features": {
                    platform_feature.name: platform_feature.supported for platform_feature in platform.features
                },
            }
            for platform in environment.platforms
        },
    }
    data["toolchains"] = environment.toolchains
    data["targets"] = {
        "default": environment.defaults.default_target,
        **{
            target.name: {
                "enabled": target.enabled,
                "features": {target_feature.name: target_feature.supported for target_feature in target.features},
            }
            for target in environment.targets
        },
    }
    data["vars"] = environment.vars
    data["flags"] = environment.flags
    return data


def write_environment_to_file(environment, filename):
    """Utility to initialize a mlonmcu environment from a YAML file."""
    if isinstance(filename, str):
        filename = pathlib.Path(filename)
    data = create_environment_dict(environment)
    with open(filename, "w") as yaml_file:
        yaml.dump(data, yaml_file, default_flow_style=False, sort_keys=False)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .tvm import _validate_tvm
from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_utvmcg(context: MlonMcuContext, params=None):
    if not _validate_tvm(context, params=params):
        return False
    return context.environment.has_backend("tvmcg")


@Tasks.provides(["utvmcg.src_dir"])
@Tasks.validate(_validate_utvmcg)
@Tasks.register(category=TaskType.BACKEND)
def clone_utvm_staticrt_codegen(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the uTVM code generator."""
    utvmcgName = utils.makeDirName("utvmcg")
    utvmcgSrcDir = context.environment.paths["deps"].path / "src" / utvmcgName
    if rebuild or not utils.is_populated(utvmcgSrcDir):
        utvmcgRepo = context.environment.repos["utvm_staticrt_codegen"]
        utils.clone_wrapper(utvmcgRepo, utvmcgSrcDir, refresh=rebuild)
    context.cache["utvmcg.src_dir"] = utvmcgSrcDir


@Tasks.needs(["utvmcg.src_dir", "tvm.src_dir"])
@Tasks.provides(["utvmcg.build_dir", "utvmcg.exe"])
@Tasks.param("dbg", False)
@Tasks.validate(_validate_utvmcg)
@Tasks.register(category=TaskType.BACKEND)
def build_utvm_staticrt_codegen(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build the uTVM code generator."""
    if not params:
        params = {}
    flags = utils.makeFlags((params["dbg"], "dbg"))
    utvmcgName = utils.makeDirName("utvmcg", flags=flags)
    utvmcgSrcDir = context.cache["utvmcg.src_dir"]
    utvmcgBuildDir = context.environment.paths["deps"].path / "build" / utvmcgName
    utvmcgInstallDir = context.environment.paths["deps"].path / "install" / utvmcgName
    utvmcgExe = utvmcgInstallDir / "compiler"
    if rebuild or not utils.is_populated(utvmcgSrcDir) or not utvmcgExe.is_file():
        utvmcgArgs = []
        utvmcgArgs.append("-DTVM_DIR=" + str(context.cache["tvm.src_dir"]))
        crtConfigPath = context.cache["tvm.src_dir"] / "apps" / "bundle_deploy" / "crt_config"
        if context:
            user_vars = context.environment.vars
            if "tvm.crt_config_dir" in user_vars:
                crtConfigPath = Path(user_vars["tvm.crt_config_dir"])
        utvmcgArgs.append("-DTVM_CRT_CONFIG_DIR=" + str(crtConfigPath))
        utils.mkdirs(utvmcgBuildDir)
        utils.cmake(
            utvmcgSrcDir,
            *utvmcgArgs,
            cwd=utvmcgBuildDir,
            debug=params["dbg"],
            live=verbose,
        )
        utils.make(cwd=utvmcgBuildDir, threads=threads, live=verbose)
        utils.mkdirs(utvmcgInstallDir)
        utils.move(utvmcgBuildDir / "utvm_staticrt_codegen", utvmcgExe)
    context.cache["utvmcg.build_dir", flags] = utvmcgBuildDir
    context.cache["utvmcg.exe", flags] = utvmcgExe
    context.export_paths.add(utvmcgExe.parent)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

from .tvmaot import TVMAOTBackend, main


class TVMAOTPlusBackend(TVMAOTBackend):
    DEFAULTS = {
        **TVMAOTBackend.DEFAULTS,
        "arena_size": 0,
        "unpacked_api": True,
        "extra_pass_config": {
            "tir.usmp.enable": True,
            "tir.usmp.algorithm": "hill_climb",
        },
    }

    name = "tvmaotplus"

    def __init__(self, runtime="crt", fmt="mlf", system_lib=False, features=None, config=None):
        super().__init__(runtime=runtime, fmt=fmt, system_lib=False, features=features, config=config)


if __name__ == "__main__":
    sys.exit(
        main(
            TVMAOTPlusBackend,
            args=sys.argv[1:],
        )
    )  # pragma: no cover



# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU RISC-V QEMU Target definitions"""

import os

import re

from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from .riscv import RISCVTarget
from .util import update_extensions

logger = get_logger()

# TODO: create (Riscv)QemuTarget with variable machine


class RiscvQemuTarget(RISCVTarget):
    """Target using a spike machine in the QEMU simulator"""

    FEATURES = RISCVTarget.FEATURES | {"vext"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "vlen": 0,  # TODO: check allowed range [128, 1024]
        "elen": 32,
        "enable_vext": False,
        "vext_spec": 1.0,
        "embedded_vext": False,
    }
    REQUIRED = RISCVTarget.REQUIRED | {"riscv32_qemu.exe"}  # TODO: 64 bit?

    def __init__(self, name="riscv_qemu", features=None, config=None):
        super().__init__(name, features=features, config=config)

    @property
    def riscv32_qemu_exe(self):
        return self.config["riscv32_qemu.exe"]

    @property
    def extensions(self):
        exts = super().extensions
        return update_extensions(exts, vext=self.enable_vext, elen=self.elen, embedded=self.embedded_vext, fpu=self.fpu)

    @property
    def attr(self):
        attrs = super().attr.split(",")
        if self.enable_vext and f"+zvl{self.vlen}b" not in attrs:
            attrs.append(f"+zvl{self.vlen}b")
        return ",".join(attrs)

    @property
    def vlen(self):
        return int(self.config["vlen"])

    @property
    def elen(self):
        return int(self.config["elen"])

    @property
    def enable_vext(self):
        value = self.config["enable_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def vext_spec(self):
        return float(self.config["vext_spec"])

    @property
    def embedded_vext(self):
        value = self.config["embedded_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def get_cpu_str(self):
        cfg = {}
        if self.enable_vext:
            cfg["v"] = "true"
            cfg["vlen"] = str(self.vlen)
            cfg["elen"] = str(self.elen)
            cfg["vext_spec"] = "v1.0"
        return ",".join([f"rv{self.xlen}"] + [f"{key}={value}" for key, value in cfg.items()])

    def get_qemu_args(self, program):
        args = []
        args.extend(["-machine", "spike"])
        args.extend(["-bios", "none"])
        args.extend(["-icount", "shift=1"])
        args.extend(["-cpu", self.get_cpu_str()])
        args.append("-nographic")
        args.extend(["-kernel", program])
        return args

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        assert len(args) == 0, "Qemu does not support passing arguments."
        qemu_args = self.get_qemu_args(program)

        if self.timeout_sec > 0:
            raise NotImplementedError
        else:
            ret = execute(
                self.riscv32_qemu_exe,
                *qemu_args,
                cwd=cwd,
                **kwargs,
            )
        return ret

    def parse_stdout(self, out, handle_exit=None):
        cpu_cycles = re.search(r"Total Cycles: (.*)", out)
        if not cpu_cycles:
            logger.warning("unexpected script output (cycles)")
            cycles = None
        else:
            cycles = int(float(cpu_cycles.group(1)))
        return cycles

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""

        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=handle_exit
            )
        total_cycles = self.parse_stdout(out, handle_exit=handle_exit)

        metrics = Metrics()
        metrics.add("Cycles", total_cycles)

        return metrics, out, []

    def get_target_system(self):
        return self.name

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        if self.enable_vext:
            major, minor = str(self.vext_spec).split(".")[:2]
            ret["RISCV_RVV_MAJOR"] = major
            ret["RISCV_RVV_MINOR"] = minor
            ret["RISCV_RVV_VLEN"] = self.vlen
        return ret


if __name__ == "__main__":
    cli(target=RiscvQemuTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for the run process."""

from mlonmcu.cli.common import kickoff_runs
from mlonmcu.cli.build import (
    handle as handle_build,
    add_build_options,
)
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.session.run import RunStage
from mlonmcu.platform.lookup import get_platforms_targets
from .helper.parse import extract_target_names, extract_platform_names, extract_config_and_feature_names


def add_compile_options(parser):
    add_build_options(parser)
    # compile_parser = parser.add_argument_group("compile options")


def get_parser(subparsers):
    """ "Define and return a subparser for the compile subcommand."""
    parser = subparsers.add_parser("compile", description="Compile model using ML on MCU flow.")
    parser.set_defaults(flow_func=handle)
    add_compile_options(parser)
    return parser


def _handle(args, context):
    handle_build(args, ctx=context)
    targets = extract_target_names(args, context=context)  # This will eventually be ignored below
    platforms = extract_platform_names(args, context=context)

    new_config, _, _, _ = extract_config_and_feature_names(args, context=context)
    platform_targets = get_platforms_targets(context, config=new_config)  # This will slow?

    assert len(context.sessions) > 0  # TODO: automatically request session if no active one is available
    session = context.sessions[-1]
    new_runs = []
    for run in session.runs:
        if run.target is None:
            # assert run.compile_platform is None
            targets_ = targets
        else:
            targets_ = [None]
        for target_name in targets_:
            new_run = run.copy()
            if target_name is not None:
                platform_name = None
                for platform in platforms:
                    candidates = platform_targets[platform]
                    if target_name in candidates:
                        platform_name = platform
                new_run.add_platform_by_name(platform_name, context=context)
                new_run.add_target_by_name(target_name, context=context)
            new_runs.append(new_run)
    session.runs = new_runs


def check_args(context, args):
    # print("CHECK ARGS")
    pass


def handle(args, ctx=None):
    if ctx:
        _handle(args, ctx)
    else:
        with MlonMcuContext(path=args.home, deps_lock="read") as context:
            _handle(args, context)
            kickoff_runs(args, RunStage.COMPILE, context)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()
Tasks = get_task_factory()

########
# ara  #
########


def _validate_ara(context: MlonMcuContext, params=None):
    return context.environment.has_target("ara")


def _validate_ara_rtl(context: MlonMcuContext, params=None):
    return context.environment.has_target("ara_rtl")


@Tasks.provides(["ara.src_dir"])
@Tasks.validate(_validate_ara)
@Tasks.register(category=TaskType.TARGET)
def clone_ara(context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()):
    """Clone the ara repository."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    flags = utils.makeFlags()
    araName = utils.makeDirName("ara", flags=flags)
    araSrcDir = context.environment.paths["deps"].path / "src" / araName
    if "ara.src_dir" in user_vars:  # TODO: also check command line flags?
        # This would overwrite the cache.ini entry which is NOT wanted! -> return false but populate gcc_name?
        araSrcDir = user_vars["ara.src_dir"]
    else:
        if rebuild or not utils.is_populated(araSrcDir):
            araRepo = context.environment.repos["ara"]
            utils.clone_wrapper(araRepo, araSrcDir, refresh=rebuild)
            utils.exec_getout("make", "apply-patches", cwd=araSrcDir / "hardware")
            utils.exec_getout("make", "bender", cwd=araSrcDir / "hardware")
    context.cache["ara.src_dir", flags] = araSrcDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from mlonmcu.target.riscv.riscv_pext_target import RVPTarget
from mlonmcu.target.riscv.riscv_vext_target import RVVTarget
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.target.riscv.util import update_extensions

from mlonmcu.logging import get_logger

from .microtvm_template_target import TemplateMicroTvmPlatformTarget

logger = get_logger()


class SpikeMicroTvmPlatformTarget(TemplateMicroTvmPlatformTarget, RVPTarget, RVVTarget):
    FEATURES = TemplateMicroTvmPlatformTarget.FEATURES | RVPTarget.FEATURES | RVVTarget.FEATURES

    DEFAULTS = {
        **TemplateMicroTvmPlatformTarget.DEFAULTS,
        **RVPTarget.DEFAULTS,
        **RVVTarget.DEFAULTS,
        "verbose": False,
        "quiet": True,
        "workspace_size_bytes": None,
        "toolchain": "gcc",
    }
    REQUIRED = (
        TemplateMicroTvmPlatformTarget.REQUIRED
        | RVPTarget.REQUIRED
        | RVVTarget.REQUIRED
        | {
            "llvm.install_dir",
            "spike.exe",
            "spike.pk",
            "microtvm_spike.src_dir",
        }
    )

    def __init__(self, name=None, features=None, config=None):
        super().__init__(name=name, features=features, config=config)
        self.template_path = self.microtvm_spike_src_dir / "template_project"
        # TODO: interate into TVM build config
        self.option_names = [
            "verbose",
            "quiet",
            "workspace_size_bytes",
            # TODO
        ]

    @property
    def microtvm_spike_src_dir(self):
        return Path(self.config["microtvm_spike.src_dir"])

    @property
    def spike_exe(self):
        return Path(self.config["spike.exe"])

    @property
    def spike_pk(self):
        return Path(self.config["spike.pk"])

    @property
    def llvm_prefix(self):
        return Path(self.config["llvm.install_dir"])

    @property
    def toolchain(self):
        value = self.config["toolchain"]
        assert value in ["gcc", "llvm"]
        return value

    def get_project_options(self):
        ret = super().get_project_options()
        ret.update(
            {
                "gcc_prefix": self.riscv_gcc_prefix,
                "gcc_name": self.riscv_gcc_basename,
                "llvm_dir": self.llvm_prefix,
                "toolchain": self.toolchain,
                "spike_exe": str(self.spike_exe),
                "spike_pk": str(self.spike_pk),
                "arch": self.arch,
                "abi": self.abi,
                "vlen": self.vlen,
                "elen": self.elen,
            }
        )
        return ret

    @property
    def extensions(self):
        # exts = RVPTarget.extensions(self) + RVVTarget.extensions(self)
        exts = super().extensions
        return update_extensions(
            exts,
        )

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        model = ret["target_model"]
        _, arch = model.split("-", 1)
        ret["target_model"] = f"spike-{arch}"
        if backend in SUPPORTED_TVM_BACKENDS:
            if optimized_layouts:
                if self.enable_pext or self.enable_vext:
                    ret.update(
                        {
                            # Warning: passing kernel layouts does not work with upstream TVM
                            # TODO: allow passing map?
                            "desired_layout": "NHWC:HWOI",
                        }
                    )
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MicroTVM Tune Platform"""
from mlonmcu.config import str2bool
from .microtvm_target_platform import MicroTvmTargetPlatform
from ..tvm.tvm_tune_platform import TvmTunePlatform


class MicroTvmTunePlatform(TvmTunePlatform, MicroTvmTargetPlatform):
    """MicroTVM Tune platform class."""

    FEATURES = TvmTunePlatform.FEATURES | MicroTvmTargetPlatform.FEATURES

    DEFAULTS = {
        **TvmTunePlatform.DEFAULTS,
        **MicroTvmTargetPlatform.DEFAULTS,
    }

    REQUIRED = TvmTunePlatform.REQUIRED | MicroTvmTargetPlatform.REQUIRED

    @property
    def experimental_tvmc_micro_tune(self):
        value = self.config["experimental_tvmc_micro_tune"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def invoke_tvmc_micro_tune(self, *args, target=None, list_options=False, **kwargs):
        all_args = []
        all_args.extend(args)
        template_args = self.get_template_args(target)
        all_args.extend(template_args)
        return self.invoke_tvmc_micro("tune", *all_args, target=target, list_options=list_options, **kwargs)

    def invoke_tvmc_tune(self, *args, target=None, **kwargs):
        return self.invoke_tvmc_micro_tune(*args, target=target, **kwargs)

    def _tune_model(self, model_path, backend, target):
        assert self.experimental_tvmc_micro_tune, "Microtvm tuning requires experimental_tvmc_micro_tune"

        return super()._tune_model(model_path, backend, target)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_vicuna(context: MlonMcuContext, params=None):
    if not context.environment.has_target("vicuna"):
        return False
    return True


@Tasks.provides(["vicuna.src_dir"])
@Tasks.validate(_validate_vicuna)
@Tasks.register(category=TaskType.TARGET)
def clone_vicuna(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the vicuna repository."""
    vicunaName = utils.makeDirName("vicuna")
    vicunaSrcDir = context.environment.paths["deps"].path / "src" / vicunaName
    user_vars = context.environment.vars
    if "vicuna.src_dir" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(vicunaSrcDir):
        vicunaRepo = context.environment.repos["vicuna"]
        utils.clone_wrapper(vicunaRepo, vicunaSrcDir, refresh=rebuild)
    context.cache["vicuna.src_dir"] = vicunaSrcDir


# REMOVE THIS AND USE VERILATOR TASK
# @Tasks.needs(["vicuna.src_dir"])
# @Tasks.provides(["vicuna.build_dir", "vicuna.install_dir", "vicuna.exe"])
# @Tasks.validate(_validate_vicuna)  # TODO: add validate_vicuna_build
# @Tasks.register(category=TaskType.TARGET)
# def build_vicuna(
#     context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
# ):
#     """Build vicuna verilator testbench."""
#     if not params:
#         params = {}
#     user_vars = context.environment.vars
#     if "vicuna.exe" in user_vars:  # TODO: also check command line flags?
#         return False
#     vicunaName = utils.makeDirName("vicuna")
#     vicunaSrcDir = context.cache["vicuna.src_dir"]
#     vicunaBuildDir = context.environment.paths["deps"].path / "build" / vicunaName
#     vicunaInstallDir = context.environment.paths["deps"].path / "install" / vicunaName
#     vicunaBin = vicunaInstallDir / "vicuna?"
#     if rebuild or not (utils.is_populated(vicunaInstallDir) and vicunaBin.is_file()):
#         utils.mkdirs(vicunaBuildDir)
#         vicunaArgs = []
#         vicunaArgs.append("--prefix=" + str(vicunaInstallDir))
#         utils.exec_getout(
#             str(vicunaSrcDir / "configure"),
#             *vicunaArgs,
#             cwd=vicunaBuildDir,
#             # live=False,
#             live=verbose,
#         )
#         utils.make(cwd=vicunaBuildDir, threads=threads, live=verbose)
#         utils.make(target="install", cwd=vicunaBuildDir, live=verbose)
#     context.cache["vicuna.build_dir"] = vicunaBuildDir
#     context.cache["vicuna.install_dir"] = vicunaInstallDir
#     context.cache["vicuna.exe"] = vicunaBin



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .common import get_task_factory

__all__ = ["get_task_factory"]

from .arm_gcc import *  # noqa: F401, F403
from .cmsisnn import *  # noqa: F401, F403
from .corstone300 import *  # noqa: F401, F403
from .espidf import *  # noqa: F401, F403
from .etiss import *  # noqa: F401, F403
from .llvm import *  # noqa: F401, F403
from .mlif import *  # noqa: F401, F403
from .muriscvnn import *  # noqa: F401, F403
from .riscv_gcc import *  # noqa: F401, F403
from .spike import *  # noqa: F401, F403
from .tflite import *  # noqa: F401, F403
from .tflmc import *  # noqa: F401, F403
from .tf import *  # noqa: F401, F403
from .tvm import *  # noqa: F401, F403
from .utvmcg import *  # noqa: F401, F403
from .zephyr import *  # noqa: F401, F403
from .pulp import *  # noqa: F401, F403
from .ara import *  # noqa: F401, F403
from .verilator import *  # noqa: F401, F403
from .ovpsim import *  # noqa: F401, F403
from .vicuna import *  # noqa: F401, F403
from .benchmarks import *  # noqa: F401, F403
from .srecord import *  # noqa: F401, F403
from .layergen import *  # noqa: F401, F403
from .dtc import *  # noqa: F401, F403
from .corev import *  # noqa: F401, F403
from .cv32e40p import *  # noqa: F401, F403



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MicroTVM Compile Platform"""
from pathlib import Path
from typing import Tuple

from mlonmcu.artifact import Artifact, ArtifactFormat

from ..platform import CompilePlatform


class MicroTvmCompilePlatform(CompilePlatform):
    """MicroTVM compile platform class."""

    def invoke_tvmc_micro_create(self, mlf_path, target=None, list_options=False, force=True, **kwargs):
        all_args = []
        if force:
            all_args.append("--force")
        all_args.append(self.project_dir)
        all_args.append(mlf_path)
        template_args = self.get_template_args(target)
        all_args.extend(template_args)
        return self.invoke_tvmc_micro("create", *all_args, target=target, list_options=list_options, **kwargs)

    def invoke_tvmc_micro_build(self, target=None, list_options=False, force=False, **kwargs):
        all_args = []
        if force:
            all_args.append("--force")
        all_args.append(self.project_dir)
        template_args = self.get_template_args(target)
        all_args.extend(template_args)
        return self.invoke_tvmc_micro("build", *all_args, target=target, list_options=list_options, **kwargs)

    def prepare(self, mlf, target):
        out = self.invoke_tvmc_micro_create(mlf, target=target)
        return out

    def compile(self, target):
        out = ""
        # TODO: build with cmake options
        out += self.invoke_tvmc_micro_build(target=target)
        return out

    def generate(self, src, target, model=None) -> Tuple[dict, dict]:
        src = Path(src) / "default.tar"  # TODO: lookup for *.tar file
        artifacts = []
        out = self.prepare(src, target)
        out += self.compile(target)
        stdout_artifact = Artifact(
            "microtvm_out.log", content=out, fmt=ArtifactFormat.TEXT  # TODO: split into one file per command
        )
        artifacts.append(stdout_artifact)
        return {"default": artifacts}, {}



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import shutil
import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_etiss(context: MlonMcuContext, params={}):
    if "dbg" in params:
        dbg = params["dbg"]
        if dbg:
            if not context.environment.has_feature("etissdbg"):
                return False
    return context.environment.has_target("etiss_pulpino") or context.environment.has_target("etiss")


def _validate_etiss_clean(context: MlonMcuContext, params={}):
    if not _validate_etiss(context, params=params):
        return False
    user_vars = context.environment.vars
    keep_build_dir = user_vars.get("etiss.keep_build_dir", True)
    return not keep_build_dir


@Tasks.provides(["etiss.src_dir"])
@Tasks.validate(_validate_etiss)
@Tasks.register(category=TaskType.TARGET)
def clone_etiss(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the ETISS repository."""
    etissName = utils.makeDirName("etiss")
    user_vars = context.environment.vars
    if "etiss.src_dir" in user_vars:
        etissSrcDir = Path(user_vars["etiss.src_dir"])
        rebuild = False
    else:
        etissSrcDir = context.environment.paths["deps"].path / "src" / etissName
    if rebuild or not utils.is_populated(etissSrcDir):
        etissRepo = context.environment.repos["etiss"]
        utils.clone_wrapper(etissRepo, etissSrcDir, refresh=rebuild)
    context.cache["etiss.src_dir"] = etissSrcDir


# @Tasks.needs(["etiss.src_dir", "llvm.install_dir"])
@Tasks.needs(["etiss.src_dir"])
@Tasks.optional(["etiss.plugins_dir"])  # Just as a dummy target to enforce order
@Tasks.provides(["etiss.build_dir", "etiss.install_dir"])
@Tasks.param("dbg", [False, True])
@Tasks.validate(_validate_etiss)
@Tasks.register(category=TaskType.TARGET)
def build_etiss(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build the ETISS simulator."""
    if not params:
        params = {}
    flags = utils.makeFlags((params["dbg"], "dbg"))
    etissName = utils.makeDirName("etiss", flags=flags)
    etissBuildDir = context.environment.paths["deps"].path / "build" / etissName
    etissInstallDir = context.environment.paths["deps"].path / "install" / etissName
    # llvmInstallDir = context.cache["llvm.install_dir"]
    user_vars = context.environment.vars
    if "etiss.build_dir" in user_vars or "etiss.install_dir" in user_vars:
        return False
    if rebuild or not utils.is_populated(etissBuildDir):
        utils.mkdirs(etissBuildDir)
        env = os.environ.copy()
        # env["LLVM_DIR"] = str(llvmInstallDir)
        utils.cmake(
            context.cache["etiss.src_dir"],
            "-DCMAKE_INSTALL_PREFIX=" + str(etissInstallDir),
            cwd=etissBuildDir,
            debug=params["dbg"],
            env=env,
            live=verbose,
        )
        utils.make(cwd=etissBuildDir, threads=threads, live=verbose)
    context.cache["etiss.build_dir", flags] = etissBuildDir
    context.cache["etiss.install_dir", flags] = etissInstallDir


@Tasks.needs(["etiss.build_dir"])
@Tasks.provides(["etiss.install_dir", "etissvp.exe", "etissvp.script"])
@Tasks.param("dbg", [False, True])
@Tasks.validate(_validate_etiss)
@Tasks.register(category=TaskType.TARGET)
def install_etiss(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Install ETISS."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    if "etiss.install_dir" in user_vars and "etissvp.exe" in user_vars and "etissvp.script" in user_vars:
        return False
    flags = utils.makeFlags((params["dbg"], "dbg"))
    # etissName = utils.makeDirName("etiss", flags=flags)
    etissBuildDir = Path(context.cache["etiss.build_dir", flags])
    etissInstallDir = Path(context.cache["etiss.install_dir", flags])
    etissvpExe = etissInstallDir / "bin" / "bare_etiss_processor"
    etissvpScript = etissInstallDir / "bin" / "run_helper.sh"
    if rebuild or not utils.is_populated(etissInstallDir) or not etissvpExe.is_file():
        utils.make("install", cwd=etissBuildDir, threads=threads, live=verbose)
    context.cache["etiss.install_dir", flags] = etissInstallDir
    context.cache["etissvp.exe", flags] = etissvpExe
    context.cache["etissvp.script", flags] = etissvpScript


@Tasks.needs(["etiss.install_dir", "etiss.build_dir"])  # TODO: make sure install has finished
@Tasks.removes(["etiss.build_dir"])  # TODO: implement
@Tasks.param("dbg", [False, True])
@Tasks.validate(_validate_etiss_clean)
@Tasks.register(category=TaskType.TARGET)
def clean_etiss(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Cleanup ETISS build dir."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    if "etiss.install_dir" in user_vars and "etissvp.exe" in user_vars and "etissvp.script" in user_vars:
        return False
    flags = utils.makeFlags((params["dbg"], "dbg"))
    # etissName = utils.makeDirName("etiss", flags=flags)
    etissBuildDir = context.cache["etiss.build_dir", flags]
    shutil.rmtree(etissBuildDir)
    del context.cache["etiss.build_dir", flags]


def _validate_microtvm_etiss(context: MlonMcuContext, params=None):
    return context.environment.has_target("microtvm_etiss")


@Tasks.provides(["microtvm_etiss.src_dir", "microtvm_etiss.template"])
@Tasks.validate(_validate_microtvm_etiss)
@Tasks.register(category=TaskType.TARGET)
def clone_microtvm_etiss(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the microtvm-etiss-template repository."""
    name = utils.makeDirName("microtvm_etiss")
    srcDir = context.environment.paths["deps"].path / "src" / name
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["microtvm_etiss"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    context.cache["microtvm_etiss.src_dir"] = srcDir
    context.cache["microtvm_etiss.template"] = srcDir / "template_project"


def _validate_etiss_accelerator_plugins(context: MlonMcuContext, params=None):
    return _validate_etiss(context, params=params) and context.environment.has_feature("vanilla_accelerator")


@Tasks.needs(["etiss.src_dir"])
@Tasks.provides(["etiss_accelerator_plugins.src_dir", "etiss.plugins_dir"])
@Tasks.validate(_validate_etiss_accelerator_plugins)
@Tasks.register(category=TaskType.FEATURE)
def clone_etiss_accelerator_plugins(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the plugins repository."""
    name = utils.makeDirName("etiss_accelerator_plugins")
    user_vars = context.environment.vars
    pluginsDir = Path(context.cache["etiss.src_dir"]) / "PluginImpl"
    if "etiss_accelerator_plugins.src_dir" in user_vars:
        srcDir = Path(user_vars["etiss_accelerator_plugins.src_dir"])
        rebuild = False
    else:
        srcDir = context.environment.paths["deps"].path / "src" / name
    # TODO: lookup directories automatically
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["etiss_accelerator_plugins"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    plugins = ["VanillaAccelerator", "QVanillaAccelerator", "QVanillaAcceleratorT"]
    for plugin in plugins:
        dest = pluginsDir / plugin
        if rebuild or not dest.is_symlink():
            if dest.is_symlink():
                dest.unlink()
            utils.symlink(srcDir / plugin, dest)
    context.cache["etiss_accelerator_plugins.src_dir"] = srcDir
    context.cache["etiss.plugins_dir"] = pluginsDir



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""UMA backend for the vanilla_accelerator accelerator"""
from .passes import VanillaAcceleratorConv2dPass, ConvertLayout
from tvm.relay.backend.contrib.uma.api.utils import PassPhase
from tvm.relay.backend.contrib.uma.backend import UMABackend
from .codegen import gen_includes
from .patterns import conv2d_pattern


class VanillaAcceleratorBackend(UMABackend):
    """UMA backend for the VanillaAccelerator accelerator."""

    print("backend")

    def __init__(self):
        super().__init__()

        # Target configuration
        self._register_target_attr("dimension")

        # Relay Pattern registration
        self._register_pattern("conv2d", conv2d_pattern())

        # Relay to Relay function registration
        self._register_relay_pass(PassPhase.PRE_PARTITIONING, ConvertLayout())

        # Relay to TIR function registration
        self._register_tir_pass(PassPhase.TIR_PHASE_0, VanillaAcceleratorConv2dPass())

        # TIR to runtime function registration
        self._register_codegen(fmt="c", includes=gen_includes)

    @property
    def target_name(self):
        return "vanilla_accelerator"



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
from argparse import Namespace
from mlonmcu.cli.helper.parse import parse_var, parse_vars, extract_feature_names, extract_config


def test_parse_var():
    assert parse_var("foo=bar") == ("foo", "bar")
    assert parse_var("foo=bar=2000") == ("foo", "bar=2000")
    # assert parse_var("foo=") == ("foo", None)


def test_parse_var_invalid():
    with pytest.raises(RuntimeError):
        parse_var("foo")
    with pytest.raises(RuntimeError):
        parse_var("=bar")


def test_parse_vars():
    assert parse_vars([]) == {}
    assert parse_vars([""]) == {}
    assert parse_vars(["foo=bar"]) == {"foo": "bar"}
    assert parse_vars(["foo=bar", "a=b"]) == {"foo": "bar", "a": "b"}


def test_extract_feature_names():
    assert extract_feature_names(Namespace(feature=None, feature_gen=None))[0] == []
    assert extract_feature_names(Namespace(feature=[], feature_gen=None))[0] == []
    assert extract_feature_names(Namespace(feature=["x"], feature_gen=None))[0] == ["x"]
    assert extract_feature_names(Namespace(feature=["x", "y"], feature_gen=None))[0] == ["x", "y"]


def test_extract_config():
    assert extract_config(Namespace(config=None, config_gen=None))[0] == {}
    assert extract_config(Namespace(config=[], config_gen=None))[0] == {}
    assert extract_config(Namespace(config=[[]], config_gen=None))[0] == {}
    assert extract_config(Namespace(config=[[], []], config_gen=None))[0] == {}
    assert extract_config(Namespace(config=[["foo=bar"]], config_gen=None))[0] == {"foo": "bar"}
    assert extract_config(Namespace(config=[["foo=bar", "a=b"]], config_gen=None))[0] == {"foo": "bar", "a": "b"}
    assert extract_config(Namespace(config=[["foo=bar"], ["a=b"]], config_gen=None))[0] == {"foo": "bar", "a": "b"}



#!/usr/bin/env python
#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import division, print_function, unicode_literals

import ttfw_idf


@ttfw_idf.idf_example_test(env_tag="Example_GENERIC", target=["esp32", "esp32s2", "esp32c3"], ci_target=["esp32"])
def test_examples_hello_world(env, extra_data):
    app_name = "hello_world"
    dut = env.get_dut(app_name, "examples/get-started/hello_world")
    dut.start_app()
    res = dut.expect(ttfw_idf.MINIMUM_FREE_HEAP_SIZE_RE)
    if not res:
        raise ValueError("Maximum heap size info not found")
    ttfw_idf.print_heap_size(app_name, dut.app.config_name, dut.TARGET, res[0])


if __name__ == "__main__":
    test_examples_hello_world()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for the run process."""

from mlonmcu.cli.common import kickoff_runs
from mlonmcu.cli.compile import (
    handle as handle_compile,
    add_compile_options,
)
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.session.run import RunStage


def add_run_options(parser):
    add_compile_options(parser)
    _ = parser.add_argument_group("run options")


def get_parser(subparsers):
    """ "Define and return a subparser for the run subcommand."""
    parser = subparsers.add_parser(
        "run",
        description="Run model using ML on MCU flow. This is meant to reproduce the behavior"
        + " of the original `run.py` script in older versions of mlonmcu.",
    )
    parser.set_defaults(flow_func=handle)
    add_run_options(parser)
    return parser


def check_args(context, args):
    # print("CHECK ARGS")
    pass


def handle(args):
    with MlonMcuContext(path=args.home, deps_lock="read") as context:
        check_args(context, args)
        handle_compile(args, context)
        kickoff_runs(args, RunStage.RUN, context)
